# file: gway/runner.py

import os
import time
import asyncio
import hashlib
import threading
import requests


# Extract all async/thread/coroutine runner logic into Runner,
# and have Gateway inherit from Runner and Resolver.
class Runner:
    """
    Runner provides async/threading/coroutine management for Gateway.
    """
    def __init__(self, *args, **kwargs):
        self._async_threads = []
        super().__init__(*args, **kwargs)

    def run_coroutine(self, func_name, coro_or_func, args=None, kwargs=None):
        try:
            start_time = time.perf_counter() if getattr(self, 'timed_enabled', False) else None
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

            if asyncio.iscoroutine(coro_or_func):
                result = loop.run_until_complete(coro_or_func)
            else:
                result = loop.run_until_complete(coro_or_func(*(args or ()), **(kwargs or {})))

            # Insert result into results if available (only if called from Gateway)
            if hasattr(self, "results"):
                self.results.insert(func_name, result)
                if isinstance(result, dict) and hasattr(self, "context"):
                    self.context.update(result)
        except Exception as e:
            if hasattr(self, "error"):
                self.error(f"Async error in {func_name}: {e}")
                if hasattr(self, "exception"):
                    self.exception(e)
        finally:
            loop.close()
            if start_time is not None:
                if hasattr(self, 'log'):
                    self.log(f"[timed] {func_name} (async) took {time.perf_counter() - start_time:.3f}s")

    def until(self, *, file=None, url=None, pypi=False, version=False, build=False,
              done=False, notify=False, notify_only=False, abort=False):
        assert file or url or pypi or version or build or done, "Use --done for unconditional looping."

        if not self._async_threads and hasattr(self, "critical"):
            self.critical("No async threads detected before entering loop.")

        from gway import gw

        abort_triggered = False
        abort_message = None

        def shutdown(reason):
            message = f"{reason} triggered async shutdown."
            if notify or notify_only:
                try:
                    gw.notify(message)
                except Exception:
                    pass
            if notify_only:
                if hasattr(self, "warning"):
                    self.warning(message + " (notify-only)")
                return
            if hasattr(self, "warning"):
                self.warning(message)
            if abort:
                nonlocal abort_triggered, abort_message
                abort_triggered = True
                abort_message = message
            self._async_threads.clear()

        watchers = []
        if version:
            watchers.append((gw.resource("VERSION"), watch_file, "VERSION file"))
        if build:
            watchers.append((gw.resource("BUILD"), watch_file, "BUILD file"))
        watchers.extend([
            (file, watch_file, "Lock file"),
            (url, watch_url, "Lock url"),
            (pypi if pypi is not False else None, watch_pypi_package, "PyPI package"),
        ])
        events = []
        for target, watcher, reason in watchers:
            if target:
                if hasattr(self, "info"):
                    self.info(f"Setup watcher for {reason}")
                if target is True and pypi:
                    target = "gway"
                events.append(watcher(target, on_change=lambda r=reason: shutdown(r)))
        try:
            while any(thread.is_alive() for thread in self._async_threads):
                time.sleep(0.1)
        except KeyboardInterrupt:
            if hasattr(self, "critical"):
                self.critical("KeyboardInterrupt received. Exiting immediately.")
            os._exit(1)
        finally:
            for e in events:
                if e:
                    e.set()
        if abort and abort_triggered:
            gw.abort(abort_message or "Async shutdown", exit_code=1)


def watch_file(*filepaths, on_change, interval=10.0, hash=False, resource=True):
    from gway import gw

    paths = []
    for path in filepaths:
        resolved = gw.resource(path) if resource else path
        if os.path.isdir(resolved):
            for root, _, files in os.walk(resolved):
                for file in files:
                    paths.append(os.path.join(root, file))
        else:
            paths.append(resolved)

    stop_event = threading.Event()

    def _watch():
        last_mtimes = {}
        last_hashes = {}

        for path in paths:
            try:
                last_mtimes[path] = os.path.getmtime(path)
                if hash:
                    with open(path, 'rb') as f:
                        last_hashes[path] = hashlib.md5(f.read()).hexdigest()
            except FileNotFoundError:
                pass

        while not stop_event.is_set():
            for path in paths:
                try:
                    current_mtime = os.path.getmtime(path)
                    if hash:
                        if path not in last_mtimes or current_mtime != last_mtimes[path]:
                            with open(path, 'rb') as f:
                                current_hash = hashlib.md5(f.read()).hexdigest()
                            if path in last_hashes and current_hash != last_hashes[path]:
                                on_change()
                                stop_event.set()
                                return
                            last_hashes[path] = current_hash
                        last_mtimes[path] = current_mtime
                    else:
                        if path in last_mtimes and current_mtime != last_mtimes[path]:
                            on_change()
                            stop_event.set()
                            return
                        last_mtimes[path] = current_mtime
                except FileNotFoundError:
                    pass
            time.sleep(interval)

    thread = threading.Thread(target=_watch, daemon=True)
    thread.start()
    return stop_event


def _retry_loop(fn, *, interval, stop_event, label):
    """Retry wrapper that logs and silently recovers from errors."""
    from gway import gw
    while not stop_event.is_set():
        try:
            fn()
        except Exception as e:
            gw.warn(f"[Watcher] {label} error: {e}")
        time.sleep(interval)


def watch_url(url, on_change, *, 
              interval=60.0, event="change", resend=False, value=None):
    stop_event = threading.Event()

    def _check():
        response = requests.get(url, timeout=5)
        content = response.content
        status_ok = 200 <= response.status_code < 400

        if event == "up":
            if status_ok:
                on_change()
                stop_event.set()
                return
        elif event == "down":
            if not status_ok:
                on_change()
                stop_event.set()
                return
        elif event == "has" and isinstance(value, str):
            if value.lower() in content.decode(errors="ignore").lower():
                on_change()
                stop_event.set()
                return
        elif event == "lacks" and isinstance(value, str):
            if value.lower() not in content.decode(errors="ignore").lower():
                on_change()
                stop_event.set()
                return
        else:  # event == "change"
            response.raise_for_status()
            nonlocal last_hash
            current_hash = hashlib.sha256(content).hexdigest()
            if last_hash is not None and current_hash != last_hash:
                on_change()
                stop_event.set()
                return
            last_hash = current_hash

    last_hash = None
    thread = threading.Thread(target=lambda: _retry_loop(
        _check, interval=interval, stop_event=stop_event, label=f"url:{url}"), daemon=True)
    thread.start()
    return stop_event


def watch_pypi_package(package_name, on_change, *, interval=3000.0):
    stop_event = threading.Event()
    url = f"https://pypi.org/pypi/{package_name}/json"

    def _check():
        nonlocal last_version
        response = requests.get(url, timeout=5)
        response.raise_for_status()
        data = response.json()
        current_version = data["info"]["version"]
        if last_version is not None and current_version != last_version:
            on_change()
            stop_event.set()
            return
        last_version = current_version

    last_version = None
    thread = threading.Thread(target=lambda: _retry_loop(
        _check, interval=interval, stop_event=stop_event, label=f"pypi:{package_name}"), daemon=True)
    thread.start()
    return stop_event



# file: projects/games/snl.py
"""Simple shared Massive Snake game (multiplayer Snakes and Ladders)."""

import json
import random
from gway import gw

BOARD_FILE = gw.resource("work", "shared", "games", "massive_snake.json", touch=True)
BOARD_SIZE = 100

# Basic snakes and ladders layout
SNAKES = {16: 6, 48: 26, 49: 11, 56: 53, 62: 19, 64: 60, 93: 73, 95: 75, 98: 78}
LADDERS = {1: 38, 4: 14, 9: 31, 21: 42, 28: 84, 36: 44, 51: 67, 71: 91, 80: 100}


def load_board():
    """Load the shared board from disk or initialize a new one."""
    if BOARD_FILE.exists() and BOARD_FILE.stat().st_size > 0:
        try:
            with open(BOARD_FILE, "r", encoding="utf-8") as f:
                return json.load(f)
        except Exception as e:
            gw.warn(f"Failed loading board: {e}")
    board = {"players": {}, "last_roll": 0}
    save_board(board)
    return board


def save_board(board):
    with open(BOARD_FILE, "w", encoding="utf-8") as f:
        json.dump(board, f)


def _use_cookies():
    return (
        hasattr(gw.web, "app")
        and hasattr(gw.web, "cookies")
        and getattr(gw.web.app, "is_setup", lambda x: False)("web.cookies")
        and gw.web.cookies.accepted()
    )


def _get_player_id():
    return gw.web.cookies.get("msnake_id") if _use_cookies() else None


def _set_player_id(pid: str):
    if _use_cookies():
        gw.web.cookies.set("msnake_id", pid, path="/", max_age=30 * 24 * 3600)


def _get_ascensions() -> int:
    if not _use_cookies():
        return 0
    try:
        return int(gw.web.cookies.get("msnake_asc") or "0")
    except Exception:
        return 0


def _set_ascensions(val: int):
    if _use_cookies():
        gw.web.cookies.set("msnake_asc", str(val), path="/", max_age=365 * 24 * 3600)


def _add_player(board, name: str, color: str) -> str:
    pid = str(random.randint(100000, 999999))
    board["players"][pid] = {"name": name, "color": color, "pos": 0}
    return pid


def _apply_move(pos: int, roll: int) -> tuple[int, str | None]:
    """Apply a dice roll and return the new position and event."""
    pos += roll
    if pos > BOARD_SIZE:
        pos = BOARD_SIZE - (pos - BOARD_SIZE)
    event = None
    if pos in SNAKES:
        pos = SNAKES[pos]
        event = "snake"
    elif pos in LADDERS:
        pos = LADDERS[pos]
        event = "ladder"
    return max(0, min(pos, BOARD_SIZE)), event


def _board_html(board: dict) -> str:
    """Return an abstract HTML board grid."""
    # collect players per position for quick lookup
    players_at = {}
    for info in board.get("players", {}).values():
        players_at.setdefault(info.get("pos", 0), []).append(info)

    rows = []
    for r in range(10):
        start = BOARD_SIZE - r * 10
        nums = range(start, start - 10, -1) if r % 2 == 0 else range(start - 9, start + 1)
        cells = []
        for n in nums:
            classes = []
            if n in SNAKES:
                classes.append("snake")
            if n in LADDERS:
                classes.append("ladder")
            marks = "".join(
                f'<span class="player" style="color:{p.get("color", "#000")}">&#9679;</span>'
                for p in players_at.get(n, [])
            )
            num = f'<div class="num">{n}</div>'
            cells.append(f'<td class="{" ".join(classes)}">{num}{marks}</td>')
        rows.append("<tr>" + "".join(cells) + "</tr>")
    return '<table class="snake-board">' + "".join(rows) + "</table>"


def view_massive_snake(*, action=None, name=None, color=None):
    """Main Massive Snake view."""
    board = load_board()
    pid = _get_player_id()
    asc = _get_ascensions()
    ready_to_ascend = False

    if not pid and name and color:
        pid = _add_player(board, name, color)
        save_board(board)
        _set_player_id(pid)

    message = ""
    roll_msg = ""
    if action == "ascend" and pid and board["players"].get(pid, {}).get("pos") >= BOARD_SIZE:
        asc += 1
        _set_ascensions(asc)
        board["players"][pid]["pos"] = 0
        save_board(board)
        message = "Ascended!"
    elif action == "roll":
        roll = random.randint(1, 6)
        board["last_roll"] = roll
        event = None
        for pid_, pdata in board["players"].items():
            new_pos, ev = _apply_move(pdata.get("pos", 0), roll)
            pdata["pos"] = new_pos
            if pid_ == pid:
                event = ev
        save_board(board)
        message = f"Rolled {roll}!"
        if event == "ladder":
            message += " You went up a ladder!"
        elif event == "snake":
            message += " You fell down a snake!"
        if pid and board["players"].get(pid, {}).get("pos", 0) >= BOARD_SIZE:
            ready_to_ascend = True
            message += " Ready to ascend!"
        roll_msg = message
        message = ""

    rows = []
    for pid_, info in board["players"].items():
        style = f" style=\"color:{info.get('color','')}\"" if info.get("color") else ""
        me = " (you)" if pid_ == pid else ""
        name_html = gw.web.nav.html_escape(info.get("name", "Player"))
        rows.append(f"<tr><td{style}>{name_html}{me}</td><td>{info.get('pos',0)}</td></tr>")

    join_form = ""
    if not pid:
        join_form = (
            "<form method='post'>"
            "<input name='name' placeholder='Name' required> "
            "<input name='color' type='color' value='#ff0000' required> "
            "<button type='submit'>Join</button>"
            "</form>"
        )

    roll_button = ""
    ascend_button = ""
    if pid:
        player_pos = board["players"].get(pid, {}).get("pos", 0)
        ready_to_ascend = ready_to_ascend or player_pos >= BOARD_SIZE
        if player_pos < BOARD_SIZE:
            roll_button = (
                "<form method='post' class='snake-roll'>"
                "<button class='snake-button roll-button' type='submit' name='action' value='roll'>Roll Dice</button>"
                f"<span class='roll-msg'>{gw.web.nav.html_escape(roll_msg)}</span>"
                "</form>"
            )
        if player_pos >= BOARD_SIZE:
            ascend_button = (
                "<form method='post' class='snake-ascend' style='display:inline-block'>"
                "<button class='snake-button ascend-button' type='submit' name='action' value='ascend'>Ascend</button>"
                "<span id='ascend-msg' class='ascend-msg'></span>"
                "</form>"
            )

    script = ""
    if ready_to_ascend and pid:
        script = (
            "<script>"
            "const title=document.getElementById('msnake-title');"
            "const msg=document.getElementById('ascend-msg');"
            "msg.textContent='Ready to ascend! ';"
            "const span=document.createElement('span');"
            "span.textContent=title.textContent;"
            "span.className='snake-title ascended';"
            "msg.appendChild(span);"
            "const roll=document.querySelector('.roll-button');"
            "const asc=document.querySelector('.snake-ascend button');"
            "asc.disabled=true; if(roll) roll.disabled=true;"
            "setTimeout(()=>{msg.remove();asc.disabled=false;if(roll) roll.disabled=false;},2000);"
            "</script>"
        )

    html = [
        '<link rel="stylesheet" href="/static/games/massive_snake/board.css">',
        "<h1 id='msnake-title' class='snake-title'>Massive Snake</h1>",
        script,
        "<p><em>A Massively Multiplayer Game of Snakes and Ladders.</em></p>",
        join_form,
        f"<p>{message}</p>" if message else "",
        f"<p>Ascensions: {asc}</p>" if pid else "",
        roll_button,
        ascend_button,
        "<table><tr><th>Player</th><th>Position</th></tr>",
        "".join(rows),
        "</table>",
        _board_html(board),
    ]
    return "\n".join(html)



import unittest
from unittest.mock import patch
from gway import gw
import sys
import os

site = gw.web.site

class FeedbackViewTests(unittest.TestCase):
    def test_feedback_form_display(self):
        with patch.dict(os.environ, {"GITHUB_TOKEN": "x"}):
            html = site.view_feedback()
            self.assertIn("<form", html)
            self.assertIn("name=\"name\"", html)
            self.assertIn("name=\"email\"", html)
            self.assertIn("name=\"topic\"", html)
            self.assertIn("name=\"message\"", html)
            self.assertIn("publicly displayed", html)
            self.assertIn("Create an Issue Report", html)

    def test_feedback_form_missing_token_with_mail(self):
        env = {
            "MAIL_SENDER": "a",
            "MAIL_PASSWORD": "b",
            "SMTP_SERVER": "smtp",
            "SMTP_PORT": "25",
        }
        with patch.dict(os.environ, env, clear=True):
            html = site.view_feedback()
            self.assertIn("<form", html)
            self.assertIn("GitHub issue creation unavailable", html)

    def test_feedback_form_missing_token_no_mail(self):
        with patch.dict(os.environ, {}, clear=True):
            html = site.view_feedback()
            self.assertIn("Feedback unavailable", html)

    def test_feedback_post_calls_issue(self):
        class FakeRequest:
            def __init__(self):
                self.method = "POST"
        with patch('bottle.request', FakeRequest()):
            with patch.dict(os.environ, {'GITHUB_TOKEN': 'x'}):
                with patch('requests.post') as p:
                    p.return_value.status_code = 201
                    p.return_value.json.return_value = {'html_url': 'http://example.com'}
                    html = site.view_feedback(name='A', email='a@example.com', topic='Test', message='Hello', create_issue=True)
                    self.assertIn('Thank you', html)
                    p.assert_called_once()
                    body = p.call_args.kwargs['json']['body']
                    self.assertNotIn('a@example.com', body)

    def test_feedback_post_without_checkbox(self):
        class FakeRequest:
            def __init__(self):
                self.method = "POST"
        with patch('bottle.request', FakeRequest()):
            with patch.dict(os.environ, {'GITHUB_TOKEN': 'x'}):
                with patch('requests.post') as p:
                    with patch.object(gw.mail, 'send') as mail_send:
                        html = site.view_feedback(name='A', email='a@example.com', topic='Test', message='Hello')
                        self.assertIn('Thank you', html)
                        p.assert_not_called()
                        mail_send.assert_called_once()

    def test_feedback_issue_failure_falls_back_to_mail(self):
        class FakeRequest:
            def __init__(self):
                self.method = "POST"
        with patch('bottle.request', FakeRequest()):
            with patch.dict(os.environ, {'GITHUB_TOKEN': 'x'}):
                with patch.object(gw.hub, 'create_issue', side_effect=RuntimeError('fail')):
                    with patch.object(gw.mail, 'send') as mail_send:
                        html = site.view_feedback(name='A', email='a@example.com', topic='Test', message='Hello', create_issue=True)
                        self.assertIn('feedback sent via email', html)
                        mail_send.assert_called_once()

    def test_feedback_issue_missing_token_falls_back_to_mail(self):
        class FakeRequest:
            def __init__(self):
                self.method = "POST"
        with patch('bottle.request', FakeRequest()):
            env = {
                "MAIL_SENDER": "a",
                "MAIL_PASSWORD": "b",
                "SMTP_SERVER": "smtp",
                "SMTP_PORT": "25",
            }
            with patch.dict(os.environ, env, clear=True):
                with patch.object(gw.hub, 'create_issue', side_effect=RuntimeError('fail')):
                    with patch.object(gw.mail, 'send') as mail_send:
                        html = site.view_feedback(name='A', email='a@example.com', topic='Test', message='Hello', create_issue=True)
                        self.assertIn('feedback sent via email', html)
                        mail_send.assert_called_once()

if __name__ == '__main__':
    unittest.main()



# file: projects/games/qpig.py
"""Prototype incremental game about quantum guinea pigs."""

from gway import gw
import time
import json
import math
import base64
from bottle import request

DEFAULT_PIGS = 1
DEFAULT_MICROCERTS = 500  # 0.5 Cert
DEFAULT_ENC_SMALL = 1
DEFAULT_ENC_LARGE = 0
DEFAULT_AVAILABLE = 3
DEFAULT_VEGGIES = {}
DEFAULT_FOOD = []

ENCLOSURE_MAX = 8

CERTAINTY_MAX = 1000  # stored in microcerts
FILL_TIME = 10 * 60  # base seconds from 0 to 1
ENC_TIME_SMALL = 5 * 60
ENC_TIME_LARGE = 8 * 60

SMALL_COST = 650
LARGE_COST = 920
UPKEEP_SMALL_HR = 2.0
UPKEEP_LARGE_HR = 3.0

ADOPTION_ADD = 2
ADOPTION_INTERVAL = 3 * 3600
ADOPTION_THRESHOLD = 7

# QP generation: 50% chance every 30s plus +/-25% from Certainty
QP_INTERVAL = 30.0  # seconds between pellet attempts
QP_BASE_CHANCE = 0.5
QP_CERT_BONUS = 0.25

VEGGIE_TYPES = ["carrot", "lettuce", "cilantro", "cucumber"]
VEGGIE_BASE_PRICE = 12
VEGGIE_PRICE_SPREAD = 8

# chance to generate an extra pellet while nibbling
VEGGIE_BONUS = {
    "carrot": 0.25,
    "lettuce": 0.15,
    "cilantro": 0.3,
    "cucumber": 0.2,
}

# how long each veggie is eaten and how long the boost lingers, in seconds
VEGGIE_EFFECTS = {
    "carrot": (60, 30),
    "lettuce": (90, 45),
    "cilantro": (120, 60),
    "cucumber": (75, 40),
}


OFFER_EXPIRY = 300  # seconds


def _load_state() -> dict:
    """Load state from request or return defaults."""
    data = request.forms.get("state") or request.query.get("state") or ""
    state = {}
    if data:
        try:
            raw = base64.b64decode(data.encode()).decode()
            state = json.loads(raw)
        except Exception:
            gw.debug("invalid state input")
    return {
        "pigs": int(state.get("pigs", DEFAULT_PIGS)),
        "mc": float(state.get("mc", DEFAULT_MICROCERTS)),
        "pellets": float(state.get("pellets", 0.0)),
        "time": float(state.get("time", time.time())),
        "avail": int(state.get("avail", DEFAULT_AVAILABLE)),
        "enc_small": int(state.get("enc_small", DEFAULT_ENC_SMALL)),
        "enc_large": int(state.get("enc_large", DEFAULT_ENC_LARGE)),
        "last_add": float(state.get("last_add", time.time())),
        "veggies": state.get("veggies", {}).copy(),
        "food": state.get("food", []).copy(),
        "offer": state.get("offer"),
    }


def _dump_state(state: dict) -> str:
    raw = json.dumps(state)
    return base64.b64encode(raw.encode()).decode()


def _get_offer(state: dict, now: float) -> tuple[str, int, int]:
    offer = state.get("offer")
    if offer and now - offer.get("time", 0) < OFFER_EXPIRY:
        return offer["kind"], int(offer["qty"]), int(offer["price"])
    import random

    kind = random.choice(VEGGIE_TYPES)
    qty = random.randint(1, 3)
    price = VEGGIE_BASE_PRICE + random.randint(-VEGGIE_PRICE_SPREAD, VEGGIE_PRICE_SPREAD)
    price = max(5, min(20, price))
    state["offer"] = {"kind": kind, "qty": qty, "price": price, "time": now}
    return kind, qty, price


def _process_state(state: dict, action: str | None = None) -> dict:
    """Update and optionally mutate the farm state."""
    gw.debug(f"_process_state start: action={action}")
    pigs = state["pigs"]
    mc = state["mc"]
    pellets = state["pellets"]
    last_time = state["time"]
    avail = state["avail"]
    enc_small = state["enc_small"]
    enc_large = state["enc_large"]
    last_add = state["last_add"]
    veggies = state["veggies"]
    food = state["food"]
    gw.debug(
        "initial state: pigs=%s mc=%s pellets=%s avail=%s enc_small=%s enc_large=%s veggies=%s food=%s",
        pigs,
        mc,
        pellets,
        avail,
        enc_small,
        enc_large,
        veggies,
        food,
    )

    now = time.time()
    dt = max(0.0, min(now - last_time, 3600))
    gw.debug(f"time delta={dt}")

    upkeep_mc = ((UPKEEP_SMALL_HR * enc_small) + (UPKEEP_LARGE_HR * enc_large)) * (
        dt / 3600
    )
    mc = max(0.0, mc - upkeep_mc)
    gw.debug(f"after upkeep: mc={mc}")

    fill_time = FILL_TIME + enc_small * ENC_TIME_SMALL + enc_large * ENC_TIME_LARGE
    cert_frac = mc / CERTAINTY_MAX
    cert_frac = 1.0 - (1.0 - cert_frac) * math.exp(-dt / fill_time)
    mc = min(CERTAINTY_MAX, cert_frac * CERTAINTY_MAX)
    gw.debug(f"after fill: mc={mc}")

    # handle food buffs
    food = [f for f in food if f[1] > now]
    buff_active = any(f[2] > now for f in food)
    buff_mult = 2.0 if buff_active else 1.0

    intervals = int(dt / QP_INTERVAL)
    if intervals > 0:
        import random
        base_prob = QP_BASE_CHANCE * (1 + (cert_frac - 0.5) * QP_CERT_BONUS * 2)
        prob = min(1.0, base_prob * buff_mult)
        gw.debug(f"pellet gen: intervals={intervals} prob={prob} pigs={pigs}")
        for _ in range(intervals):
            for _ in range(pigs):
                if random.random() < prob:
                    pellets += 1
                    if random.random() < prob:
                        pellets += 1
        last_time += intervals * QP_INTERVAL

    # chance for bonus pellets from active veggies
    nibbling = [f[0] for f in food if f[1] > now]
    if nibbling:
        import random
        gw.debug(f"nibbling: {nibbling}")
        for kind in nibbling:
            chance = VEGGIE_BONUS.get(kind, 0.0)
            if random.random() < chance:
                pellets += 1

    # replenish adoption queue
    while now - last_add >= ADOPTION_INTERVAL and avail <= ADOPTION_THRESHOLD:
        avail = min(avail + ADOPTION_ADD, ADOPTION_THRESHOLD)
        last_add += ADOPTION_INTERVAL
        gw.debug(f"adoption queue replenished -> avail={avail}")

    offer_kind, offer_qty, offer_price = _get_offer(state, now)

    capacity = enc_small + enc_large * 2
    if action == "adopt" and avail > 0 and pigs < capacity:
        pigs += 1
        avail -= 1
        gw.debug(f"action adopt -> pigs={pigs} avail={avail}")
    elif action == "buy_small" and mc >= SMALL_COST and (enc_small + enc_large) < ENCLOSURE_MAX:
        mc -= SMALL_COST
        enc_small += 1
        gw.debug(f"action buy_small -> enc_small={enc_small} mc={mc}")
    elif action == "buy_large" and mc >= LARGE_COST and (enc_small + enc_large) < ENCLOSURE_MAX:
        mc -= LARGE_COST
        enc_large += 1
        gw.debug(f"action buy_large -> enc_large={enc_large} mc={mc}")
    elif action == "buy_veggie":
        total = offer_price * offer_qty
        if mc >= total:
            mc -= total
            veggies[offer_kind] = veggies.get(offer_kind, 0) + offer_qty
            offer_kind, offer_qty, offer_price = _get_offer(state, now)
            gw.debug(f"action buy_veggie -> veggies={veggies} mc={mc}")
    elif action and action.startswith("place_"):
        kind = action.split("_", 1)[1]
        slots = (enc_small + enc_large) * 3
        if veggies.get(kind, 0) > 0 and len(food) < slots:
            veggies[kind] -= 1
            if veggies[kind] == 0:
                del veggies[kind]
            eat, linger = VEGGIE_EFFECTS.get(kind, (60, 30))
            food.append([kind, now + eat, now + eat + linger])
            gw.debug(f"action place_{kind} -> food={food} veggies={veggies}")

    state.update(
        {
            "pigs": pigs,
            "mc": mc,
            "pellets": pellets,
            "time": last_time,
            "avail": avail,
            "enc_small": enc_small,
            "enc_large": enc_large,
            "last_add": last_add,
            "veggies": veggies,
            "food": food,
            "offer": {
                "kind": offer_kind,
                "qty": offer_qty,
                "price": offer_price,
                "time": state.get("offer", {}).get("time", now),
            },
        }
    )
    gw.debug(f"_process_state end -> {state}")
    return state


def render_qpig_farm_stats(state: dict) -> str:
    """Return the farm stats block."""
    gw.debug("render_qpig_farm_stats called")
    pigs = state["pigs"]
    mc = state["mc"]
    pellets = state["pellets"]
    avail = state["avail"]
    enc_small = state["enc_small"]
    enc_large = state["enc_large"]
    veggies = state["veggies"]
    food = state["food"]
    capacity = enc_small + enc_large * 2
    veg_list = ", ".join(f"{k}:{v}" for k, v in veggies.items()) or "none"
    food_list = ", ".join(f"{f[0]}" for f in food) or "none"
    html = [
        f"<p>You have <b>{pigs}</b> quantum pigs (capacity {capacity}).</p>",
        f"<p>Certainty: <b>{int(mc)}</b> MC / {CERTAINTY_MAX}</p>",
        f"<p>QPellets: <b>{int(pellets)}</b></p>",
        f"<p>Veggies: {veg_list}</p>",
        f"<p>Food placed: {food_list}</p>",
        f"<p>Pigs waiting for adoption: {avail}</p>",
    ]
    return "\n".join(html)


def view_qpig_farm(*, action: str = None, **_):
    """Main Quantum Piggy farm view."""
    gw.debug(f"view_qpig_farm called with action={action}")
    state = _load_state()
    state = _process_state(state, action)
    state_b64 = _dump_state(state)

    offer = state["offer"]
    offer_kind, offer_qty, offer_price = offer["kind"], offer["qty"], offer["price"]
    capacity = state["enc_small"] + state["enc_large"] * 2

    html = [
        '<link rel="stylesheet" href="/static/games/qpig/farm.css">',
        '<div class="qpig-garden">',
        "<h1>Quantum Piggy Farm</h1>",
        "<canvas id='qpig-canvas' width='32' height='32'></canvas>",
        '<div id="qpig-stats">',
        render_qpig_farm_stats(state),
        "</div>",
        "<form method='post' id='qpig-form' class='qpig-buttons'>",
        "<input type='hidden' name='state' id='qpig-state'>",
    ]
    if state["avail"] > 0 and state["pigs"] < capacity:
        html.append("<button type='submit' name='action' value='adopt'>Adopt Pig</button>")
    html.extend(
        [
            f"<button type='submit' name='action' value='buy_small'>Buy Enclosure ({SMALL_COST} MC)</button>",
            f"<button type='submit' name='action' value='buy_large'>Buy Large Enclosure ({LARGE_COST} MC)</button>",
            f"<button type='submit' name='action' value='buy_veggie'>Buy {offer_qty} {offer_kind}(s) ({offer_price * offer_qty} MC)</button>",
        ]
    )

    slots = (state["enc_small"] + state["enc_large"]) * 3
    if len(state["food"]) < slots:
        for k, v in state["veggies"].items():
            if v > 0:
                html.append(f"<button type='submit' name='action' value='place_{k}'>Feed {k}</button>")

    html.extend([
        "<button type='button' id='qpig-save' title='Save'>\ud83d\udcbe</button>",
        "<button type='button' id='qpig-load' title='Load'>\ud83d\udcc2</button>",
        "</form>",
        "</div>",
    ])

    script = """
<script>
const KEY='qpig_state';
sessionStorage.setItem(KEY, '{state_b64}');
const form=document.getElementById('qpig-form');
const hidden=document.getElementById('qpig-state');
if(form){{form.addEventListener('submit',()=>{{hidden.value=sessionStorage.getItem(KEY)||'';}});}}
const save=document.getElementById('qpig-save');
if(save){{save.addEventListener('click',()=>{{const data=sessionStorage.getItem(KEY)||'';const blob=new Blob([data],{{type:'application/octet-stream'}});const a=document.createElement('a');a.href=URL.createObjectURL(blob);a.download='qpig-save.qpg';a.click();setTimeout(()=>URL.revokeObjectURL(a.href),1000);}});}}
const load=document.getElementById('qpig-load');
if(load){{load.addEventListener('click',()=>{{const inp=document.createElement('input');inp.type='file';inp.accept='.qpg';inp.onchange=e=>{{const f=e.target.files[0];if(!f)return;const r=new FileReader();r.onload=ev=>{{sessionStorage.setItem(KEY, ev.target.result.trim());location.reload();}};r.readAsText(f);}};inp.click();}});}}
const canvas=document.getElementById('qpig-canvas');
if(canvas){{const ctx=canvas.getContext('2d');const img=new Image();img.src='/static/games/qpig/pig.png';img.onload=()=>{{ctx.imageSmoothingEnabled=false;ctx.drawImage(img,0,0);}};}}
</script>
""".format(state_b64=state_b64)

    html.append(script)
    return "\n".join(html)



# file: projects/web/static.py

import os
from gway import gw

def collect(*, css="global", js="global", root="data/static", target="work/shared"):
    enabled = getattr(gw.web.app, "enabled_projects", lambda: set())()
    static_root = gw.resource(root)

    def find_files(kind, proj):
        found = []
        seen = set()
        parts = proj.split('.')
        # Recursively walk project path
        if parts:
            proj_path = os.path.join(static_root, *parts)
            for rootdir, dirs, files in os.walk(proj_path):
                rel_root = os.path.relpath(rootdir, static_root)
                for fname in files:
                    if kind == "css" and fname.endswith(".css"):
                        rel = os.path.join(rel_root, fname)
                    elif kind == "js" and fname.endswith(".js"):
                        rel = os.path.join(rel_root, fname)
                    else:
                        continue
                    if rel not in seen:
                        seen.add(rel)
                        found.append((proj, rel, os.path.join(rootdir, fname)))
        # Ancestors, only direct files
        for i in range(len(parts)-1, -1, -1):
            ancestor_path = os.path.join(static_root, *parts[:i])
            if not os.path.isdir(ancestor_path):
                continue
            rel_ancestor = os.path.relpath(ancestor_path, static_root)
            for fname in os.listdir(ancestor_path):
                fpath = os.path.join(ancestor_path, fname)
                if not os.path.isfile(fpath):
                    continue
                if kind == "css" and fname.endswith(".css"):
                    rel = os.path.join(rel_ancestor, fname) if rel_ancestor != "." else fname
                elif kind == "js" and fname.endswith(".js"):
                    rel = os.path.join(rel_ancestor, fname) if rel_ancestor != "." else fname
                else:
                    continue
                if rel not in seen:
                    seen.add(rel)
                    found.append((proj, rel, fpath))
        return found

    report = {"css": [], "js": []}
    # --- Collect CSS ---
    if css:
        all_css = []
        for proj in enabled:
            all_css.extend(find_files("css", proj))
        seen_css = set()
        for entry in all_css:
            if entry[1] not in seen_css:
                seen_css.add(entry[1])
                report["css"].append(entry)
        if isinstance(css, str):
            bundle_path = gw.resource(target, f"{css}.css")
            with open(bundle_path, "w", encoding="utf-8") as out:
                for proj, rel, full in reversed(report["css"]):
                    with open(full, "r", encoding="utf-8") as f:
                        out.write(f"/* --- {proj}:{rel} --- */\n")
                        out.write(f.read() + "\n\n")
            report["css_bundle"] = bundle_path

    # --- Collect JS ---
    if js:
        all_js = []
        for proj in enabled:
            all_js.extend(find_files("js", proj))
        seen_js = set()
        for entry in all_js:
            if entry[1] not in seen_js:
                seen_js.add(entry[1])
                report["js"].append(entry)
        if isinstance(js, str):
            bundle_path = gw.resource(target, f"{js}.js")
            with open(bundle_path, "w", encoding="utf-8") as out:
                for proj, rel, full in report["js"]:
                    with open(full, "r", encoding="utf-8") as f:
                        out.write(f"// --- {proj}:{rel} ---\n")
                        out.write(f.read() + "\n\n")
            report["js_bundle"] = bundle_path

    return report



# file: projects/ocpp/sink.py

import json
import traceback
from fastapi import FastAPI, WebSocket, WebSocketDisconnect
from gway import gw


def setup_sink_app(*, app=None):
    """
    Basic OCPP passive sink for messages, acting as a dummy CSMS server.
    This won't pass compliance or provide authentication. It just accepts and logs all.
    Note: This version of the app was tested at the EVCS with real EVs.
    """
    # A - This line ensures we find just the kind of app we need or create one if missing
    oapp = app
    match app:
        case FastAPI() as f:
            app = f
            _is_new_app = False
        case list() | tuple() as seq:
            app = next((x for x in seq if isinstance(x, FastAPI)), None)
            _is_new_app = app is None
        case None:
            _is_new_app = True
        case _ if isinstance(app, FastAPI):
            _is_new_app = False
        case _ if hasattr(app, "__iter__") and not isinstance(app, (str, bytes, bytearray)):
            app = next((x for x in app if isinstance(x, FastAPI)), None)
            _is_new_app = app is None
        case _:
            _is_new_app = app is None or not isinstance(app, FastAPI)
    if _is_new_app:
        app = FastAPI()

    @app.websocket("{path:path}")
    async def websocket_ocpp(websocket: WebSocket, path: str):
        gw.info(f"[OCPP] New WebSocket connection at /{path}")
        try:
            await websocket.accept()
            while True:
                raw = await websocket.receive_text()
                gw.info(f"[OCPP:{path}] Message received:", raw)

                try:
                    msg = json.loads(raw)
                    if isinstance(msg, list) and len(msg) >= 3 and msg[0] == 2:
                        message_id = msg[1]
                        action = msg[2]
                        payload = msg[3] if len(msg) > 3 else {}

                        gw.info(f"[OCPP:{path}] -> Action: {action} | Payload: {payload}")
                        response = [3, message_id, {"status": "Accepted"}]
                        await websocket.send_text(json.dumps(response))
                        gw.info(f"[OCPP:{path}] <- Acknowledged: {response}")
                    else:
                        gw.warning(f"[OCPP:{path}] Received non-Call message or malformed")
                except Exception as e:
                    gw.error(f"[OCPP:{path}] Error parsing message: {e}")
                    gw.debug(traceback.format_exc())

        except WebSocketDisconnect:
            gw.info(f"[OCPP:{path}] Disconnected")
        except Exception as e:
            gw.error(f"[OCPP:{path}] WebSocket error: {e}")
            gw.debug(traceback.format_exc())

    # B- This return pattern ensures we include our app in the bundle (if any)
    if _is_new_app:
        return app if not oapp else (oapp, app)    
    return oapp



# file: projects/web/dj.py
"""Helpers to install and control an embedded Django project."""

import os
import subprocess
import sys
from pathlib import Path

from gway import gw

DEFAULT_DIR = gw.resource("work", "web", "dj")


def _run(cmd, *, cwd=None):
    """Execute a command list, logging and streaming output."""
    gw.info(f"Running: {' '.join(cmd)} (cwd={cwd})")
    process = subprocess.run(cmd, cwd=cwd, check=False)
    if process.returncode != 0:
        gw.error(f"Command {cmd[0]} exited with code {process.returncode}")
    return process.returncode


def install(*, target: str | Path = DEFAULT_DIR, version: str | None = None, project: str = "config"):
    """Install Django and start a project in the target directory.

    Parameters:
        target: Path where the project lives. Defaults to ``work/web/dj``.
        version: Optional version specifier passed to ``pip install``.
        project: Name for the Django project folder (default ``config``).
    """
    target = Path(target)
    os.makedirs(target, exist_ok=True)
    spec = "django" + (version or "")
    _run([sys.executable, "-m", "pip", "install", spec])
    if not (target / "manage.py").exists():
        _run(["django-admin", "startproject", project, "."], cwd=target)
    return str(target)


def manage(*args, path: str | Path = DEFAULT_DIR):
    """Run ``manage.py`` command in the environment."""
    path = Path(path)
    cmd = [sys.executable, "manage.py", *args]
    return _run(cmd, cwd=path)


def start(*, path: str | Path = DEFAULT_DIR, addrport: str = "127.0.0.1:8000"):
    """Start the Django development server."""
    return manage("runserver", addrport, path=path)


def stop(*, pattern: str = "manage.py", signal: str = "TERM"):
    """Stop the Django development server using pkill if available."""
    return _run(["pkill", f"-{signal}", "-f", pattern])

def view_dj(*, action: str = None, host: str = "127.0.0.1", port: int = 8000):
    """Install and control the embedded Django server."""
    project_dir = Path(DEFAULT_DIR)
    manage_py = project_dir / "manage.py"

    if action == "install":
        install(target=project_dir)
        return "<p>Django installed.</p>"

    if action == "start":
        if not manage_py.exists():
            install(target=project_dir)
        start(path=project_dir, addrport=f"{host}:{port}")
        url = f"http://{host}:{port}/"
        return f"<p>Django server started at <a href='{url}'>{url}</a></p>"

    if action == "stop":
        stop()
        return "<p>Django server stopped.</p>"

    if not manage_py.exists():
        return (
            "<p>Django not installed. Use ?action=install to install it</p>"
        )

    return (
        "<h1>Django</h1><p>Use ?action=start or ?action=stop to control the "
        "local server in work/web/dj.</p>"
    )



# projects/cdv.py

import os
from gway import gw
from urllib.parse import quote, unquote


def _encode(val):
    # Only encode if non-empty string
    return quote(val, safe='') if isinstance(val, str) and val else val


def _decode(val):
    return unquote(val) if isinstance(val, str) and val else val


def _resolve_path(pathlike):
    """Resolve to absolute path using gw.resource."""
    return gw.resource(pathlike)


def _read_table(path):
    """Read and parse a CDV table file."""
    if not path:
        return {}
    if not os.path.exists(path):
        gw.error(f"Table file not found: {path}")
        return {}
    result = {}
    with open(path, "r") as f:
        for lineno, line in enumerate(f, start=1):
            line = line.strip()
            if not line or ":" not in line:
                continue
            parts = line.split(":")
            entry = parts[0].strip()
            fields = {}
            for part in parts[1:]:
                if "=" in part:
                    k, v = part.split("=", 1)
                    fields[k.strip()] = _decode(v.strip())
            result[entry] = fields
    return result


def _write_table(path, records):
    """Write the complete dict of records to a CDV table file."""
    with open(path, "w") as f:
        for entry_id, fields in records.items():
            line = entry_id + "".join(
                f":{k}={_encode(v)}" for k, v in fields.items()
            )
            f.write(line + "\n")


def load_all(pathlike: str) -> dict[str, dict[str, str]]:
    """Load CDV table with ID followed by colon-separated key=value fields."""
    try:
        path = _resolve_path(pathlike)
        return _read_table(path)
    except Exception as e:
        gw.halt(f"Failed to read table '{pathlike}': {e}")


def update(table_path: str, entry_id: str, **fields):
    """Append or update a record in the CDV table, preserving unspecified fields."""
    if not entry_id or not table_path:
        return
    path = _resolve_path(table_path)
    records = _read_table(path)
    existing = records.get(entry_id, {})
    existing.update(fields)
    records[entry_id] = existing
    _write_table(path, records)
    gw.info(f"Updated table with ID={entry_id}")


def validate(table_path: str, entry: str, *, validator=None) -> bool:
    """
    Validate a CDV entry by ID directly from file.
    Always reloads the file from disk to avoid stale data.
    """
    path = _resolve_path(table_path)
    table = _read_table(path)
    if not table:
        gw.warn("No table loaded — rejecting validation request.")
        return False
    record = table.get(entry)
    if not record:
        return False
    if validator:
        try:
            return bool(validator(**record))
        except Exception as e:
            gw.error(f"validator failed: {e}")
            return False
    return True


def copy(table_path: str, old_entry: str, new_entry: str, **kwargs) -> bool:
    """Copy a record from old_entry to new_entry, optionally updating fields."""
    path = _resolve_path(table_path)
    records = _read_table(path)
    if old_entry not in records:
        gw.warn(f"Entry '{old_entry}' does not exist; cannot copy.")
        return False
    # Shallow copy of fields
    new_fields = records[old_entry].copy()
    new_fields.update(kwargs)
    records[new_entry] = new_fields
    _write_table(path, records)
    gw.info(f"Copied '{old_entry}' to '{new_entry}' with updates: {kwargs}")
    return True


def move(table_path: str, old_entry: str, new_entry: str, **kwargs) -> bool:
    """Move a record from old_entry to new_entry, optionally updating fields."""
    path = _resolve_path(table_path)
    records = _read_table(path)
    if old_entry not in records:
        gw.warn(f"Entry '{old_entry}' does not exist; cannot move.")
        return False
    new_fields = records[old_entry].copy()
    new_fields.update(kwargs)
    records[new_entry] = new_fields
    del records[old_entry]
    _write_table(path, records)
    gw.info(f"Moved '{old_entry}' to '{new_entry}' with updates: {kwargs}")
    return True


def credit(table_path: str, entry: str, *, field: str = 'balance', **kwargs) -> bool:
    """Add 1 (or amount from kwargs) to the given field for a record."""
    path = _resolve_path(table_path)
    records = _read_table(path)
    if entry not in records:
        gw.warn(f"Entry '{entry}' does not exist; cannot credit.")
        return False
    try:
        amt = float(kwargs.pop('amount', 1))
        prev = float(records[entry].get(field, 0))
        records[entry][field] = str(prev + amt)
        records[entry].update(kwargs)
        _write_table(path, records)
        gw.info(f"Credited {amt} to '{entry}' field '{field}'. New value: {records[entry][field]}")
        return True
    except Exception as e:
        gw.error(f"credit failed: {e}")
        return False


def debit(table_path: str, entry: str, *, field: str = 'balance', **kwargs) -> bool:
    """Subtract 1 (or amount from kwargs) from the given field for a record."""
    path = _resolve_path(table_path)
    records = _read_table(path)
    if entry not in records:
        gw.warn(f"Entry '{entry}' does not exist; cannot debit.")
        return False
    try:
        amt = float(kwargs.pop('amount', 1))
        prev = float(records[entry].get(field, 0))
        records[entry][field] = str(prev - amt)
        records[entry].update(kwargs)
        _write_table(path, records)
        gw.info(f"Debited {amt} from '{entry}' field '{field}'. New value: {records[entry][field]}")
        return True
    except Exception as e:
        gw.error(f"debit failed: {e}")
        return False


def view_colon_validator(*, text=None):
    
    # can paste a CDV file into a large textarea and submit for validation as 'text'.
    # If text is received, test and validate it and let the user know if there are any issues
    # found that would impede the file to be processed as a valid CDV. 

    raise NotImplementedError


def save_all(pathlike: str, all_records: dict[str, dict[str, str]]):
    """
    Replace all records in the CDV file at pathlike with the given dict.
    """
    path = _resolve_path(pathlike)
    _write_table(path, all_records)


def read_rows(pathlike: str) -> list[list[str]]:
    """
    Read a CDV as a list of rows: [id, k1, v1, ...].
    """
    records = load_all(pathlike)
    rows = []
    for entry_id, fields in records.items():
        row = [entry_id]
        for k, v in fields.items():
            row += [k, v]
        rows.append(row)
    return rows


def write_rows(pathlike: str, rows: list[list[str]]):
    """
    Write a list of rows: [id, k1, v1, ...] as a CDV file.
    """
    recs = {}
    for row in rows:
        if not row: continue
        entry_id = row[0]
        fields = {row[i]: row[i+1] for i in range(1, len(row)-1, 2)}
        recs[entry_id] = fields
    save_all(pathlike, recs)


def delete(table_path: str, entry_id: str):
    """
    Remove a record by ID from the CDV table.
    """
    path = _resolve_path(table_path)
    records = _read_table(path)
    if entry_id in records:
        del records[entry_id]
        _write_table(path, records)
        gw.info(f"Deleted record '{entry_id}' from table.")
        return True
    return False


def _sanitize_cdv_path(pathlike: str) -> str:
    """Return a sanitized path string safe for gw.resource."""
    from projects.web.site import _sanitize_filename

    parts = [
        _sanitize_filename(p)
        for p in str(pathlike).replace("\\", "/").split("/")
        if p and p not in {"..", "."}
    ]
    return "/".join(parts)


def _parse_cdv_text(text: str) -> dict[str, dict[str, str]]:
    """Parse colon-delimited text into a record dict."""
    records = {}
    for line in text.splitlines():
        line = line.strip()
        if not line or ":" not in line:
            continue
        parts = line.split(":")
        entry_id = parts[0].strip()
        fields = {}
        for part in parts[1:]:
            if "=" in part:
                k, v = part.split("=", 1)
                fields[k.strip()] = _decode(v.strip())
        if entry_id:
            records[entry_id] = fields
    return records


def _records_to_text(records: dict[str, dict[str, str]]) -> str:
    """Serialize record dict back to colon-delimited text."""
    lines = []
    for entry_id, fields in records.items():
        line = entry_id + "".join(
            f":{k}={_encode(v)}" for k, v in fields.items()
        )
        lines.append(line)
    return "\n".join(lines)


def view_data_editor(*, cdv: str = None, text: str = None):
    """Interactive editor for CDV files."""
    from bottle import request, response
    import html

    if not cdv:
        return (
            "<h1>CDV Data Editor</h1>"
            "<form method='get'>"
            "<input name='cdv' placeholder='path/to/file.cdv' required> "
            "<button type='submit'>Open</button>"
            "</form>"
        )

    cdv_safe = _sanitize_cdv_path(cdv)
    try:
        path = gw.resource(*cdv_safe.split("/"))
    except Exception as e:
        return f"<p>Invalid CDV path: {html.escape(str(e))}</p>"

    if request.method == "POST":
        records = _parse_cdv_text(text or "")
        save_all(path, records)
        response.status = 303
        url = gw.web.app.build_url("data-editor", cdv=cdv_safe)
        response.set_header("Location", url)
        return ""

    records = load_all(path)
    cdv_text = html.escape(_records_to_text(records))
    return (
        f"<h1>Editing {html.escape(cdv_safe)}</h1>"
        "<form method='post'>"
        f"<textarea name='text' rows='15' style='width:100%;'>{cdv_text}</textarea>"
        f"<input type='hidden' name='cdv' value='{html.escape(cdv_safe)}'>"
        "<button type='submit'>Save</button>"
        "</form>"
    )



import unittest
from gway import gw

class ProjectVerbFallbackTests(unittest.TestCase):
    def test_single_word_function_fallback(self):
        app_direct = gw.web.app.setup_app("dummy")
        app_alias = gw.web.app.setup("dummy")
        self.assertEqual(type(app_alias), type(app_direct))

if __name__ == "__main__":
    unittest.main()



# file: projects/monitor/nmcli.py

"""
GWAY NMCLI Network Monitor Project

Network monitoring helpers for Linux systems using ``nmcli``.
All state is read/written via ``gw.monitor.get_state/set_states('nmcli', {...})``.
The monitors only collect information and never modify the network.

Monitors:
    - monitor_nmcli: Collect information about all interfaces.
    - monitor_ap_only: Record wlan0 status without changes.
    - monitor_station_only: Same as above for station mode.

Renders:
    - render_nmcli: Main network diagnostic report (HTML).
    - render_status: Short summary/status indicator.
    - render_monitor: Fallback renderer.
"""

import subprocess
import shlex
import html
from bottle import request
from gway import gw
from gway.sigils import _unquote

def now_iso():
    import datetime
    return datetime.datetime.now().isoformat(timespec="seconds")

# --- Utility functions ---

def nmcli(*args):
    result = subprocess.run(["nmcli", *args], capture_output=True, text=True)
    return result.stdout.strip()

def nmcli_list_connections():
    """Return a list of (name, uuid, type, device) tuples from nmcli."""
    output = nmcli('-t', '-f', 'NAME,UUID,TYPE,DEVICE', 'connection', 'show')
    conns = []
    for line in output.splitlines():
        if not line:
            continue
        parts = line.split(':')
        if len(parts) < 4:
            parts += [''] * (4 - len(parts))
        name, uuid, ctype, device = parts[:4]
        conns.append((name, uuid, ctype, device))
    return conns

def _sanitize(val):
    return _unquote(val.strip()) if isinstance(val, str) else val

def get_wlan_ifaces():
    output = nmcli("device", "status")
    wlans = []
    for line in output.splitlines():
        if line.startswith("wlan"):
            name = line.split()[0]
            if name != "wlan0":
                wlans.append(name)
    return wlans

def get_eth0_ip():
    output = nmcli("device", "show", "eth0")
    for line in output.splitlines():
        if "IP4.ADDRESS" in line:
            return line.split(":")[-1].strip()
    return None

def get_device_info(dev):
    """
    Returns a dict with relevant info for a network device, such as state, type, driver, path, mac, etc.
    """
    info = {
        'device': dev,
        'type': '-',
        'state': '-',
        'driver': '-',
        'mac': '-',
        'path': '-',
        'connection': '-',
    }
    try:
        output = nmcli('device', 'show', dev)
        for line in output.splitlines():
            if line.startswith('GENERAL.TYPE'):
                info['type'] = line.split(':', 1)[-1].strip()
            elif line.startswith('GENERAL.STATE'):
                info['state'] = line.split(':', 1)[-1].strip()
            elif line.startswith('GENERAL.HWADDR'):
                info['mac'] = line.split(':', 1)[-1].strip()
            elif line.startswith('GENERAL.DRIVER'):
                info['driver'] = line.split(':', 1)[-1].strip()
            elif line.startswith('GENERAL.PATH'):
                info['path'] = line.split(':', 1)[-1].strip()
            elif line.startswith('GENERAL.CONNECTION'):
                info['connection'] = line.split(':', 1)[-1].strip()
    except Exception as e:
        info['error'] = str(e)
    return info

def get_all_devices():
    """
    Returns a list of all device names from nmcli (regardless of status).
    """
    devices = []
    out = nmcli('device', 'status')
    for line in out.splitlines():
        parts = line.split()
        if parts:
            devices.append(parts[0])
    return devices

def get_default_route_iface():
    """Return the interface used for the default route, or None."""
    try:
        out = subprocess.check_output(["ip", "route", "show", "default"], text=True)
        for line in out.splitlines():
            if line.startswith("default"):
                parts = line.split()
                if "dev" in parts:
                    idx = parts.index("dev")
                    if idx + 1 < len(parts):
                        return parts[idx + 1]
    except Exception:
        pass
    return None

def ping(iface, target="8.8.8.8", count=2, timeout=2):
    try:
        result = subprocess.run(
            ["ping", "-I", iface, "-c", str(count), "-W", str(timeout), target],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        return result.returncode == 0
    except Exception:
        return False

def get_wlan_status(iface):
    output = nmcli("device", "status")
    for line in output.splitlines():
        if line.startswith(iface):
            fields = line.split()
            conn = (fields[2] == "connected")
            # Try to get SSID (from nmcli device show iface)
            ssid = None
            info = nmcli("device", "show", iface)
            for inf in info.splitlines():
                if "GENERAL.CONNECTION" in inf:
                    conn_name = inf.split(":")[-1].strip()
                    if conn_name and conn_name != "--":
                        # Try nmcli connection show <name> for ssid
                        det = nmcli("connection", "show", conn_name)
                        for dline in det.splitlines():
                            if "802-11-wireless.ssid" in dline:
                                ssid = dline.split(":")[-1].strip()
                                break
            inet = ping(iface)
            status = {"ssid": ssid, "connected": conn, "inet": inet}
            gw.debug(f"[nmcli] status for {iface}: {status}")
            return status
    return {"ssid": None, "connected": False, "inet": False}

def gather_eth0_status():
    """Record eth0 IP and whether a default gateway exists."""
    try:
        routes = subprocess.check_output(["ip", "route", "show", "dev", "eth0"], text=True)
        ip_addr = get_eth0_ip()
        state_update = {
            "eth0_ip": ip_addr,
            "eth0_gateway": "default" in routes,
        }
        gw.monitor.set_states('nmcli', state_update)
    except Exception as e:
        gw.monitor.set_states('nmcli', {"last_error": f"eth0 status: {e}"})

def ap_profile_exists(ap_con, ap_ssid, ap_password):
    for name, uuid, ctype, device in nmcli_list_connections():
        if name == ap_con and ctype == "wifi":
            details = nmcli("connection", "show", name)
            details_dict = {}
            for detline in details.splitlines():
                if ':' in detline:
                    k, v = detline.split(':', 1)
                    details_dict[k.strip()] = v.strip()
            ssid_ok = (details_dict.get("802-11-wireless.ssid") == ap_ssid)
            pwd_ok  = (not ap_password or details_dict.get("802-11-wireless-security.psk") == ap_password)
            return ssid_ok and pwd_ok
    return False

def ensure_ap_profile(ap_con, ap_ssid, ap_password, ap_ip=None):
    ap_con = _sanitize(ap_con)
    ap_ssid = _sanitize(ap_ssid)
    ap_password = _sanitize(ap_password)
    if not ap_con:
        raise ValueError("AP_CON must be specified.")
    if not ap_ssid or not ap_password:
        gw.info("[nmcli] Missing AP_SSID or AP_PASSWORD. Skipping AP profile creation.")
        return
    if ap_profile_exists(ap_con, ap_ssid, ap_password):
        return
    conns = nmcli("connection", "show")
    for line in conns.splitlines():
        if line.startswith(ap_con + " "):
            gw.info(f"[nmcli] Removing existing AP connection profile: {ap_con}")
            nmcli("connection", "down", ap_con)
            nmcli("connection", "delete", ap_con)
            break
    gw.info(f"[nmcli] Creating AP profile: name={ap_con} ssid={ap_ssid}")
    nmcli("connection", "add", "type", "wifi", "ifname", "wlan0",
          "con-name", ap_con, "autoconnect", "no", "ssid", ap_ssid)

    local_ip = _sanitize(gw.resolve('[AP_GATEWAY]', default=ap_ip or '10.42.0.1'))
    mod_args = [
        "mode", "ap", "802-11-wireless.band", "bg",
        "wifi-sec.key-mgmt", "wpa-psk",
        "wifi-sec.psk", ap_password,
        "ipv4.method", "shared",
    ]
    if local_ip:
        mod_args += ["ipv4.addresses", f"{local_ip}/24"]
    nmcli("connection", "modify", ap_con, *mod_args)
    

def set_wlan0_ap(ap_con, ap_ssid, ap_password):
    ap_con = _sanitize(ap_con)
    ap_ssid = _sanitize(ap_ssid)
    ap_password = _sanitize(ap_password)
    ensure_ap_profile(ap_con, ap_ssid, ap_password)
    gw.info(f"[nmcli] Activating wlan0 AP: conn={ap_con}, ssid={ap_ssid}")
    nmcli("device", "disconnect", "wlan0")
    nmcli("connection", "up", ap_con)
    gw.monitor.set_states('nmcli', {
        "wlan0_mode": "ap",
        "wlan0_ssid": ap_ssid,
        "last_config_change": now_iso(),
        "last_config_action": f"Activated AP {ap_ssid}"
    })

def set_wlan0_station():
    gw.info("[nmcli] Setting wlan0 to station (managed) mode")
    nmcli("device", "set", "wlan0", "managed", "yes")
    nmcli("device", "disconnect", "wlan0")
    gw.monitor.set_states('nmcli', {
        "wlan0_mode": "station",
        "last_config_change": now_iso(),
        "last_config_action": "Set wlan0 to station"
    })

def maybe_notify_ap_switch(ap_ssid, email=None):
    state = gw.monitor.get_state('nmcli')
    prev_mode = state.get("wlan0_mode")
    prev_ssid = state.get("wlan0_ssid")
    prev_inet = state.get("wlan0_inet")
    recipient = email if email else gw.resolve('[ADMIN_EMAIL]', default=None)
    if not recipient:
        return
    if prev_mode == "station" and prev_inet:
        subject = "[nmcli] wlan0 switching to AP mode"
        body = (
            f"Previous mode: station\n"
            f"SSID: {prev_ssid}\n"
            f"Internet: {prev_inet}\n\n"
            f"New mode: ap\n"
            f"AP SSID: {ap_ssid}\n"
        )
        try:
            gw.mail.send(subject, body=body, to=recipient)
        except Exception as e:
            gw.error(f"[nmcli] Email notification failed: {e}")

def clean_and_reconnect_wifi(iface, ssid, password=None):
    gw.debug(f"[nmcli] clean_and_reconnect_wifi({iface}, ssid={ssid})")
    for name, uuid, conn_type, device in nmcli_list_connections():
        if conn_type == "wifi" and (device == iface or name == ssid):
            gw.info(f"[nmcli] Removing stale connection {name} ({uuid}) on {iface}")
            nmcli("connection", "down", name)
            nmcli("connection", "delete", name)
            gw.monitor.set_states('nmcli', {
                "last_config_change": now_iso(),
                "last_config_action": f"Removed stale WiFi {name} on {iface}"
            })
            break
    gw.info(f"[nmcli] Resetting interface {iface}")
    nmcli("device", "disconnect", iface)
    nmcli("device", "set", iface, "managed", "yes")
    subprocess.run(["ip", "addr", "flush", "dev", iface])
    subprocess.run(["dhclient", "-r", iface])
    gw.info(f"[nmcli] Re-adding {iface} to SSID '{ssid}'")
    if password:
        nmcli("device", "wifi", "connect", ssid, "ifname", iface, "password", password)
    else:
        nmcli("device", "wifi", "connect", ssid, "ifname", iface)
    gw.monitor.set_states('nmcli', {
        "last_config_change": now_iso(),
        "last_config_action": f"Re-added {iface} to {ssid}"
    })

def try_connect_wlan0_known_networks():
    """Try connecting wlan0 using known WiFi profiles.

    Returns the SSID if connection succeeds, otherwise None.
    """
    wifi_conns = [name for name, _, ctype, _ in nmcli_list_connections() if ctype == "wifi"]
    gw.debug(f"[nmcli] known wifi profiles: {wifi_conns}")
    for conn in wifi_conns:
        gw.info(f"[nmcli] Trying wlan0 connect: {conn}")
        nmcli("device", "wifi", "connect", conn, "ifname", "wlan0")
        if ping("wlan0"):
            gw.info(f"[nmcli] wlan0 internet works via {conn}")
            gw.monitor.set_states('nmcli', {
                "wlan0_mode": "station",
                "wlan0_ssid": conn,
                "wlan0_inet": True,
                "last_config_change": now_iso(),
                "last_config_action": f"wlan0 connected to {conn}"
            })
            return conn
        clean_and_reconnect_wifi("wlan0", conn)
        gw.debug(f"[nmcli] retrying connection to {conn} after reset")
        if ping("wlan0"):
            gw.info(f"[nmcli] wlan0 internet works via {conn} after reset")
            gw.monitor.set_states('nmcli', {
                "wlan0_mode": "station",
                "wlan0_ssid": conn,
                "wlan0_inet": True,
                "last_config_change": now_iso(),
                "last_config_action": f"wlan0 reconnected to {conn}"
            })
            return conn
    gw.monitor.set_states('nmcli', {"wlan0_inet": False})
    return None

# --- Main single-run monitor functions ---

def monitor_nmcli(**kwargs):
    gather_eth0_status()
    wlan_ifaces = get_wlan_ifaces()
    gw.info(f"[nmcli] WLAN ifaces detected: {wlan_ifaces}")
    wlanN = {}
    found_inet = False
    internet_iface = None
    internet_ssid = None
    for iface in wlan_ifaces:
        s = get_wlan_status(iface)
        wlanN[iface] = s
        gw.info(f"[nmcli] {iface} status: {s}")
        if s["inet"] and not found_inet:
            found_inet = True
            internet_iface = iface
            internet_ssid = s.get("ssid")
    gw.monitor.set_states('nmcli', {"wlanN": wlanN})
    gw_iface = None
    if not found_inet:
        gw_iface = get_default_route_iface()
        gw.debug(f"[nmcli] default route iface: {gw_iface}")
        if gw_iface and ping(gw_iface):
            found_inet = True
            internet_iface = gw_iface
            if gw_iface in wlanN:
                internet_ssid = wlanN[gw_iface].get("ssid")

    if not internet_iface and gw_iface == "wlan0":
        internet_ssid = gw.monitor.get_state('nmcli').get("wlan0_ssid")

    gw.monitor.set_states('nmcli', {
        "last_monitor_check": now_iso(),
        "internet_iface": internet_iface,
        "internet_ssid": internet_ssid,
    })
    gw.debug(
        f"[nmcli] monitor results: found_inet={found_inet}, "
        f"internet_iface={internet_iface}, internet_ssid={internet_ssid}"
    )
    state = gw.monitor.get_state('nmcli')
    return {
        "ok": found_inet,
        "action": state.get("last_config_action"),
        "wlan0_mode": state.get("wlan0_mode"),
    }

def monitor_ap_only(**kwargs):
    """Record wlan0 status without changing configuration."""
    gather_eth0_status()
    status = get_wlan_status("wlan0")
    gw.monitor.set_states('nmcli', {
        "last_monitor_check": now_iso(),
        "wlanN": {"wlan0": status},
    })
    return {"wlan0_mode": gw.monitor.get_state('nmcli').get("wlan0_mode"), "ssid": status.get("ssid")}

def monitor_station_only(**kwargs):
    """Record wlan0 status without changing configuration."""
    gather_eth0_status()
    status = get_wlan_status("wlan0")
    gw.monitor.set_states('nmcli', {
        "last_monitor_check": now_iso(),
        "wlanN": {"wlan0": status},
    })
    return {"wlan0_mode": gw.monitor.get_state('nmcli').get("wlan0_mode")}

# --- Renderers (for dashboard, html output) ---

def _color_icon(status):
    if status is True or status == "ok":
        return '<span style="color:#0b0;">&#9679;</span>'
    if status is False or status == "fail":
        return '<span style="color:#b00;">&#9679;</span>'
    return '<span style="color:#bb0;">&#9679;</span>'


def render_nmcli():
    s = gw.monitor.get_state('nmcli')
    wlanN = s.get("wlanN") or {}
    internet_iface = s.get("internet_iface")
    internet_ssid = s.get("internet_ssid")

    # Fallback detection from wlan statuses
    if not internet_iface:
        for iface, st in wlanN.items():
            if st.get('inet'):
                internet_iface = iface
                internet_ssid = st.get('ssid')
                break
    if not internet_iface and s.get('wlan0_inet'):
        internet_iface = 'wlan0'
        internet_ssid = s.get('wlan0_ssid')

    # Gather device info for eth0, wlan0, all wlanN, and any other network devices
    devices = get_all_devices()
    device_info = {dev: get_device_info(dev) for dev in devices}
    wlan_count = len([d for d in devices if d.startswith('wlan') and d != 'wlan0'])

    html = ['<div class="nmcli-report">']
    html.append("<h2>Network Manager</h2>")
    html.append(f"<b>Last monitor check:</b> {s.get('last_monitor_check') or '-'}<br>")
    last_action = s.get('last_config_action')
    last_change = s.get('last_config_change')
    if last_action and last_change:
        html.append(f"<b>Last action:</b> {last_change} - {last_action}<br>")
    else:
        html.append(f"<b>Last action:</b> {last_action or '-'}<br>")
    html.append(f"<b>wlan0 mode:</b> {s.get('wlan0_mode') or '-'}<br>")
    # AP info
    wlan0_info = device_info.get('wlan0', {})
    html.append(
        f"<b>wlan0 ssid:</b> {s.get('wlan0_ssid') or '-'} "
        f"(state: {wlan0_info.get('state','-')}, driver: {wlan0_info.get('driver','-')}, "
        f"mac: {wlan0_info.get('mac','-')})<br>"
    )
    html.append(f"<b>wlan0 internet:</b> {_color_icon(s.get('wlan0_inet'))} {s.get('wlan0_inet')}<br>")

    # eth0 info
    eth0_ip = s.get('eth0_ip')
    eth0_color = _color_icon(bool(eth0_ip))
    eth0_info = device_info.get('eth0', {})
    html.append(
        f"<b>eth0 IP:</b> {eth0_color} {eth0_ip or '-'} "
        f"(state: {eth0_info.get('state','-')}, driver: {eth0_info.get('driver','-')}, "
        f"mac: {eth0_info.get('mac','-')})<br>"
    )
    eth0_gw = s.get('eth0_gateway')
    html.append(f"<b>eth0 gateway:</b> {_color_icon(eth0_gw)} {'yes' if eth0_gw else 'no'}<br>")

    html.append(f"<b>Last internet OK:</b> {_color_icon(bool(s.get('last_inet_ok')))} {s.get('last_inet_ok') or '-'}<br>")
    html.append(f"<b>Last internet fail:</b> {_color_icon(bool(s.get('last_inet_fail')))} {s.get('last_inet_fail') or '-'}<br>")
    html.append(f"<b>Last error:</b> {_color_icon(s.get('last_error') is None)} {s.get('last_error') or '-'}<br>")

    # All wlanN and relevant info (including disconnected/disabled)
    html.append(f"<b>WLANN interfaces:</b> {wlan_count}<br>")
    html.append('<table style="border-collapse:collapse;margin-top:4px;font-size:90%;">'
                '<tr>'
                '<th style="border:1px solid #ccc;padding:2px 4px;">iface</th>'
                '<th style="border:1px solid #ccc;padding:2px 4px;">SSID</th>'
                '<th style="border:1px solid #ccc;padding:2px 4px;">Connected</th>'
                '<th style="border:1px solid #ccc;padding:2px 4px;">INET</th>'
                '<th style="border:1px solid #ccc;padding:2px 4px;">State</th>'
                '<th style="border:1px solid #ccc;padding:2px 4px;">Driver</th>'
                '<th style="border:1px solid #ccc;padding:2px 4px;">MAC</th>'
                '</tr>')
    for dev in sorted([d for d in devices if d.startswith('wlan')]):
        st = wlanN.get(dev, {})
        dinfo = device_info.get(dev, {})
        gw_mark = ' <b>(gw)</b>' if dev == internet_iface else ''
        html.append(
            f"<tr>"
            f'<td style="border:1px solid #ccc;padding:2px 4px;">{dev}</td>'
            f'<td style="border:1px solid #ccc;padding:2px 4px;">{st.get("ssid") or "-"}</td>'
            f'<td style="border:1px solid #ccc;padding:2px 4px;">{_color_icon(st.get("connected"))} {st.get("connected") if "connected" in st else dinfo.get("state", "-")}</td>'
            f'<td style="border:1px solid #ccc;padding:2px 4px;">{_color_icon(st.get("inet")) if "inet" in st else _color_icon(dinfo.get("state")=="connected")} {st.get("inet") if "inet" in st else "-"}{gw_mark}</td>'
            f'<td style="border:1px solid #ccc;padding:2px 4px;">{dinfo.get("state", "-")}</td>'
            f'<td style="border:1px solid #ccc;padding:2px 4px;">{dinfo.get("driver", "-")}</td>'
            f'<td style="border:1px solid #ccc;padding:2px 4px;">{dinfo.get("mac", "-")}</td>'
            f"</tr>"
        )
    html.append('</table>')

    # Internet gateway info (with device detail)
    if internet_iface:
        gwdev = device_info.get(internet_iface, {})
        html.append(
            f"<b>Internet via:</b> {internet_iface} (SSID: {internet_ssid}) "
            f"(driver: {gwdev.get('driver','-')}, mac: {gwdev.get('mac','-')}, state: {gwdev.get('state','-')})<br>"
        )
    else:
        html.append(f"<b>Internet via:</b> <span style='color:#b00;'>No gateway detected</span><br>")

    # --- Command box ---
    html.append(_render_run_form())

    html.append("</div>")
    return "\n".join(html)


def _render_run_form(cmd: str = "", output: str = "") -> str:
    url = gw.web.app.build_url("run") if hasattr(gw, "web") else "run"
    form = [
        f"<form method='post' action='{url}' style='margin-top:8px;'>",
        f"<input type='text' name='cmd' value='{html.escape(cmd, quote=True)}' placeholder='nmcli arguments' style='width:70%;'>",
        "<button type='submit'>Run</button>",
        "</form>",
    ]
    if output:
        form.append(f"<pre>{html.escape(output)}</pre>")
    return "".join(form)


def view_get_run(cmd: str = ""):
    """Display nmcli command form."""
    return _render_run_form(cmd)


def view_post_run(cmd: str = ""):
    """Execute nmcli command and show output."""
    cmd = cmd or request.forms.get("cmd", "")
    output = ""
    if cmd:
        try:
            output = nmcli(*shlex.split(cmd))
        except Exception as e:
            output = f"Error: {e}"
    return _render_run_form(cmd, output)



# file: projects/web/chat/whats.py
"""Minimal WhatsApp Cloud API helpers."""

import os
import requests
from gway import gw


def send_message(to: str, body: str, *, token: str | None = None, phone_id: str | None = None, preview_url: bool = False):
    """Send a text message using the WhatsApp Cloud API.

    Parameters
    ----------
    to : str
        Destination phone number in international format ("+123...").
    body : str
        Text body to send.
    token : str, optional
        Access token. Defaults to ``WHATS_TOKEN`` environment variable.
    phone_id : str, optional
        Business phone number ID. Defaults to ``WHATS_PHONE_ID`` environment variable.
    preview_url : bool, optional
        Enable link preview in message body.

    Returns
    -------
    dict
        Parsed JSON response or ``{"error": ...}`` on failure.
    """
    token = token or os.environ.get("WHATS_TOKEN")
    phone_id = phone_id or os.environ.get("WHATS_PHONE_ID")
    if not token or not phone_id:
        gw.error("WhatsApp credentials missing.")
        return {"error": "Missing WhatsApp token or phone id."}

    url = f"https://graph.facebook.com/v17.0/{phone_id}/messages"
    headers = {"Authorization": f"Bearer {token}", "Content-Type": "application/json"}
    payload = {
        "messaging_product": "whatsapp",
        "to": to,
        "type": "text",
        "text": {"body": gw.resolve(body), "preview_url": bool(preview_url)},
    }
    try:
        resp = requests.post(url, json=payload, headers=headers, timeout=30)
        return resp.json()
    except Exception as e:
        gw.error(f"WhatsApp send error: {e}")
        return {"error": str(e)}


def api_post_send(*, to=None, body=None, preview_url=False, token=None, phone_id=None, **kwargs):
    """POST /chat/whats/send - Send a WhatsApp message."""
    if not (to and body):
        return {"error": "'to' and 'body' are required"}
    return send_message(str(to), str(body), token=token, phone_id=phone_id, preview_url=preview_url)



import unittest
import importlib.util
from pathlib import Path
from unittest import mock

class FakeRequest:
    def __init__(self):
        self.cookies = {}
    def get_cookie(self, name, default=None):
        return self.cookies.get(name, default)

class FakeResponse:
    def __init__(self):
        self.set_calls = []
    def set_cookie(self, name, value, **kwargs):
        self.set_calls.append((name, value, kwargs))

class CookiesUtilTests(unittest.TestCase):
    @staticmethod
    def _load_cookies():
        cookies_path = Path(__file__).resolve().parents[1] / "projects" / "web" / "cookies.py"
        spec = importlib.util.spec_from_file_location("webcookies", cookies_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    @classmethod
    def setUpClass(cls):
        cls.cookies = cls._load_cookies()

    def setUp(self):
        self.request = FakeRequest()
        self.response = FakeResponse()
        self.preq = mock.patch.object(self.cookies, 'request', self.request)
        self.pres = mock.patch.object(self.cookies, 'response', self.response)
        self.preq.start()
        self.pres.start()

    def tearDown(self):
        self.preq.stop()
        self.pres.stop()

    def test_set_and_get_when_accepted(self):
        self.request.cookies['cookies_accepted'] = 'yes'
        self.cookies.set('foo', 'bar')
        self.assertEqual(len(self.response.set_calls), 1)
        name, value, _ = self.response.set_calls[0]
        self.assertEqual((name, value), ('foo', 'bar'))
        # get should read from request
        self.request.cookies['foo'] = 'bar'
        self.assertEqual(self.cookies.get('foo'), 'bar')
        self.assertIsNone(self.cookies.get('missing'))
        self.assertTrue(self.cookies.accepted())

    def test_set_ignored_when_not_accepted(self):
        self.cookies.set('foo', 'bar')
        self.assertEqual(self.response.set_calls, [])
        # cookies_accepted cookie can always be set
        self.cookies.set('cookies_accepted', 'yes')
        self.assertEqual(len(self.response.set_calls), 1)
        self.assertEqual(self.response.set_calls[0][0], 'cookies_accepted')

    def test_remove_when_accepted(self):
        self.request.cookies['cookies_accepted'] = 'yes'
        self.cookies.remove('foo')
        self.assertEqual(len(self.response.set_calls), 2)
        for call in self.response.set_calls:
            name, value, params = call
            self.assertEqual(name, 'foo')
            self.assertEqual(value, '')
            self.assertEqual(params['expires'], 'Thu, 01 Jan 1970 00:00:00 GMT')
        self.assertFalse(self.response.set_calls[0][2]['secure'])
        self.assertTrue(self.response.set_calls[1][2]['secure'])

    def test_remove_noop_when_not_accepted(self):
        self.cookies.remove('foo')
        self.assertEqual(self.response.set_calls, [])

    def test_append_when_accepted(self):
        self.request.cookies['cookies_accepted'] = 'yes'
        self.request.cookies['bag'] = 'a=1|b=2'
        result = self.cookies.append('bag', 'c', '3')
        self.assertEqual(result, ['a=1', 'b=2', 'c=3'])
        self.assertEqual(self.response.set_calls[-1][0], 'bag')
        self.assertEqual(self.response.set_calls[-1][1], 'a=1|b=2|c=3')

    def test_append_noop_when_not_accepted(self):
        self.request.cookies['bag'] = 'a=1|b=2'
        result = self.cookies.append('bag', 'c', '3')
        self.assertEqual(result, [])
        self.assertEqual(self.response.set_calls, [])

    def test_accepted_checks_cookie_value(self):
        self.request.cookies['cookies_accepted'] = 'yes'
        self.assertTrue(self.cookies.accepted())
        self.request.cookies['cookies_accepted'] = 'no'
        self.assertFalse(self.cookies.accepted())
        del self.request.cookies['cookies_accepted']
        self.assertFalse(self.cookies.accepted())


if __name__ == '__main__':
    unittest.main()



# This is a collection of functions which I consider to be the
# collective public interface of GWAY. One of this should be the
# right entry-point depending on what channel you're comming from.

from .gateway import Gateway, gw
from .console import cli_main, process, load_recipe
from .sigils import Sigil, Resolver, Spool, __
from .structs import Results
from .logging import setup_logging
from .envs import load_env



# file: projects/web/site.py

import os
import html
import shlex
import traceback
from docutils.core import publish_parts
from pathlib import Path
import re
from gway import gw, __
from gway.console import process, chunk
import markdown as mdlib

_DEFAULT_TOME = __('[README]', 'README')

def view_reader(
    *parts,
    tome=_DEFAULT_TOME,
    ext=None,
    origin="root",
    **kwargs,
):
    """
    Render a resource file (.rst or .md) as HTML.
    If origin='root', only files in the resource root (no subfolders).
    Never serves files starting with dot or underscore.
    """
    if parts and (tome == _DEFAULT_TOME or tome == "README"):
        tome = "/".join(str(p).strip("/") for p in parts)

    gw.verbose(f"[reader] Called with tome={tome!r}, ext={ext!r}, origin={origin!r}")
    fname = _sanitize_filename(tome)
    gw.verbose(f"[reader] Sanitized filename: {fname}")

    ext = (str(ext).strip().lower() if ext else None)
    if ext and ext.startswith('.'):
        ext = ext[1:]
    gw.verbose(f"[reader] Normalized ext: {ext}")

    # Security: Never allow files starting with dot/underscore
    if _is_hidden_or_private(fname):
        gw.verbose(f"[reader] Access denied due to hidden/private filename: {fname}")
        return "<b>Access denied.</b>"

    def file_variants(base):
        if base.endswith('/'):
            base = base.rstrip('/') + '/README'
        if ext in {'rst', 'md'}:
            variants = [f"{base}.{ext}"]
        else:
            base_, ext_ = os.path.splitext(base)
            if ext_ in {'.rst', '.md'}:
                variants = [base]
            else:
                variants = [f"{base}.rst", f"{base}.md"]
        gw.verbose(f"[reader] Candidate variants for base {base}: {variants}")
        return variants

    use_root = origin == "root" and "/" not in tome
    if use_root:
        resource_dir = os.path.dirname(gw.resource('README.rst'))  # This IS the resource root
        gw.verbose(f"[reader] Resource root directory: {resource_dir}")
        for candidate in file_variants(fname):
            gw.verbose(f"[reader] Checking candidate: {candidate}")
            if _is_hidden_or_private(candidate):
                gw.verbose(f"[reader] Skipped hidden/private candidate: {candidate}")
                continue
            try:
                resource_path = gw.resource(candidate)
                gw.verbose(f"[reader] Resolved resource path: {resource_path}")
            except Exception as e:
                gw.verbose(f"[reader] gw.resource({candidate}) exception: {e}")
                continue
            # Only allow files in the resource root
            if not os.path.isfile(resource_path):
                gw.verbose(f"[reader] Not a file: {resource_path}")
                continue
            if os.path.dirname(resource_path) != resource_dir:
                gw.verbose(f"[reader] File not in resource root: {resource_path}")
                continue
            if _is_hidden_or_private(os.path.basename(resource_path)):
                gw.verbose(f"[reader] File rejected (hidden/private): {resource_path}")
                continue
            gw.verbose(f"[reader] Will open file: {resource_path}")
            try:
                with open(resource_path, encoding="utf-8") as f:
                    content = f.read()
                if candidate.lower().endswith(".rst"):
                    html = publish_parts(source=content, writer_name="html")["html_body"]
                elif candidate.lower().endswith(".md"):
                    html = mdlib.markdown(content)
                else:
                    gw.verbose(f"[reader] Unsupported file type for {candidate}")
                    html = "<b>Unsupported file type.</b>"
                gw.verbose(f"[reader] Successfully rendered {candidate}")
                return html
            except Exception as e:
                gw.verbose(f"[reader] Exception reading or rendering {resource_path}: {e}")
                continue
        exts = ' or '.join(['.rst', '.md']) if not ext else f".{ext}"
        gw.verbose(f"[reader] File not found or not allowed: {fname}{exts}; falling back to static")

    if origin == "root":
        origin = "static"

    if origin == "static":
        base_dir = Path(gw.resource('data', 'static'))
        safe_path = _sanitize_relpath(tome)
        gw.verbose(f"[reader] Static safe path: {safe_path!r}")
        if not safe_path:
            return "<b>Access denied.</b>"
        candidate_dir = Path(base_dir, safe_path)
        if candidate_dir.is_dir():
            safe_path = f"{safe_path.rstrip('/')}/README"
        for candidate in file_variants(safe_path):
            gw.verbose(f"[reader] Checking static candidate: {candidate}")
            parts = candidate.split('/')
            resource_path = gw.resource('data', 'static', *parts)
            resolved = Path(resource_path).resolve()
            if not resolved.is_file() or base_dir not in resolved.parents:
                gw.verbose(f"[reader] Invalid static path: {resolved}")
                continue
            if any(_is_hidden_or_private(p) for p in parts):
                gw.verbose(f"[reader] Hidden/private segment in {candidate}")
                continue
            try:
                with open(resolved, encoding='utf-8') as f:
                    content = f.read()
                if resolved.suffix == '.rst':
                    html = publish_parts(source=content, writer_name='html')['html_body']
                elif resolved.suffix == '.md':
                    html = mdlib.markdown(content)
                else:
                    html = '<b>Unsupported file type.</b>'
                return html
            except Exception as e:
                gw.verbose(f"[reader] Exception reading {resolved}: {e}")
                continue
        exts = ' or '.join(['.rst', '.md']) if not ext else f".{ext}"
        return f"<b>File not found or not allowed: {safe_path}{exts}</b>"

    # Fallback for other origins (not fully implemented here)
    gw.verbose(f"[view_reader] Non-root origin {origin} not implemented in this snippet.")
    return "<b>Invalid or unsupported origin.</b>"

def _sanitize_filename(fname):
    """
    Sanitize the filename: only allow dots and alphanumerics. No slashes, no backslashes, no "..".
    """
    fname = str(fname)
    fname = fname.replace('/', '').replace('\\', '').replace('..', '')
    fname = ''.join(c for c in fname if c.isalnum() or c in '._-')
    return fname

def _sanitize_relpath(path):
    """Sanitize a relative path under data/static."""
    parts = []
    segments = str(path).split('/')
    for i, segment in enumerate(segments):
        segment = segment.strip()
        if not segment:
            if i == len(segments) - 1:
                continue  # allow trailing slash
            return None
        if segment.startswith('.') or segment.startswith('_') or '..' in segment:
            return None
        clean = ''.join(c for c in segment if c.isalnum() or c in '._-')
        parts.append(clean)
    return '/'.join(parts)

def _is_hidden_or_private(fname):
    """
    Returns True if filename (or its extension) starts with a dot or underscore.
    """
    fname = os.path.basename(fname)
    if not fname:
        return True
    if fname[0] in {'.', '_'}:
        return True
    name, ext = os.path.splitext(fname)
    if ext and ext[1:2] in {'.', '_'}:
        return True
    if name.startswith('.') or name.startswith('_'):
        return True
    return False

def _looks_like_html(text: str) -> bool:
    """Heuristic check for HTML content."""
    if not isinstance(text, str):
        return False
    stripped = text.strip()
    if not stripped:
        return False
    return bool(re.search(r'<[a-zA-Z][^>]*>', stripped))

def view_help(topic="", *args, **kwargs):
    """
    Render dynamic help based on GWAY introspection and search-style links.
    If there is an exact match in the search, show it at the top (highlighted).
    """

    # gw.gelp at all times, compliment this result with other information.

    topic_in = topic or ""

    # --- Local console commands via search box ---
    if topic_in.strip().startswith(">"):
        if gw.web.server.is_local():
            cmd_str = topic_in.strip()[1:].strip()
            if not cmd_str:
                return "<i>No command provided.</i>"
            import shlex, html
            from gway.console import chunk, process
            try:
                tokens = shlex.split(cmd_str)
                commands = chunk(tokens)
                results, _ = process(commands)
                html_parts = []
                for r in results:
                    if r is None:
                        continue
                    if isinstance(r, str) and _looks_like_html(r):
                        html_parts.append(r)
                    else:
                        html_parts.append(gw.cast.to_html(r))
                if not html_parts:
                    res_dict = gw.results.get_results()
                    res_html = gw.cast.to_html(res_dict)
                    return (
                        "<div class='cli-result'>"
                        "<i>No result returned. Showing gw.results:</i><hr>"
                        + res_html + "</div>"
                    )
                return "<div class='cli-result'>" + "<hr>".join(html_parts) + "</div>"
            except Exception as ex:
                if gw.debug_enabled:
                    tb = traceback.format_exc()
                    log_tail = ""
                    try:
                        log_path = gw.resource("logs", "gway.log")
                        with open(log_path) as lf:
                            log_tail = "".join(lf.readlines()[-20:])
                    except Exception:
                        log_tail = "(unable to read log)"
                    return (
                        "<h2>Command Error</h2>"
                        f"<pre>{html.escape(str(ex))}</pre>"
                        f"<pre>{html.escape(tb)}</pre>"
                        f"<pre>{html.escape(log_tail)}</pre>"
                    )
                return f"<pre>{html.escape(str(ex))}</pre>"
        else:
            return "<b>Console commands disabled (not local).</b>"
          
    topic = topic.replace(" ", "/").replace(".", "/").replace("-", "_") if topic else ""
    parts = [p for p in topic.strip("/").split("/") if p]

    if not parts:
        help_info = gw.help()
        title = "Available Projects"
        content = "<ul>"
        for project in help_info["Available Projects"]:
            content += f'<li><a href="?topic={project}">{project}</a></li>'
        content += "</ul>"
        return f"<h1>{title}</h1>{content}"

    elif len(parts) == 1:
        project = parts[0]
        help_info = gw.help(project)
        title = f"Help Topics for <code>{project}</code>"

    else:
        *project_path, maybe_function = parts
        obj = gw
        for segment in project_path:
            obj = getattr(obj, segment, None)
            if obj is None:
                return f"<h2>Not Found</h2><p>Project path invalid at <code>{segment}</code>.</p>"
        project_str = ".".join(project_path)
        if hasattr(obj, maybe_function):
            function = maybe_function
            help_info = gw.help(project_str, function, full=True)
            full_name = f"{project_str}.{function}"
            title = f"Help for <code>{full_name}</code>"
        else:
            help_info = gw.help(project_str)
            full_name = f"{project_str}.{maybe_function}"
            title = f"Help Topics for <code>{full_name}</code>"

    if help_info is None:
        return "<h2>Not Found</h2><p>No help found for the given input.</p>"

    highlight_js = '''
    <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.9.0/styles/github-dark.min.css">
    <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.9.0/highlight.min.js"></script>
    <script>
      window.addEventListener('DOMContentLoaded',function(){
        if(window.hljs){
          document.querySelectorAll('pre code.python').forEach(el => { hljs.highlightElement(el); });
        }
      });
    </script>
    '''

    # --- Exact match highlighting logic ---
    # Only applies if help_info contains "Matches"
    if "Matches" in help_info:
        matches = help_info["Matches"]
        exact_key = (topic_in.replace(" ", "/").replace(".", "/").replace("-", "_")).strip("/")
        # Try to find an exact match (project, or project/function) in matches
        def canonical_str(m):
            p, f = m.get("Project", ""), m.get("Function", "")
            return (f"{p}/{f}" if f else p).replace(".", "/").replace("-", "_")
        exact = None
        exact_idx = -1
        for idx, m in enumerate(matches):
            if canonical_str(m).lower() == exact_key.lower():
                exact = m
                exact_idx = idx
                break

        sections = []
        # If found, show exact at top with highlight
        if exact is not None:
            sections.append('<div class="help-exact">' + _render_help_section(exact, use_query_links=True, highlight=True) + '</div>')
            # Add separator if there are more matches
            if len(matches) > 1:
                sections.append('<hr class="help-sep">')
            # Remove exact match from below
            rest = [m for i, m in enumerate(matches) if i != exact_idx]
        else:
            rest = matches

        for idx, match in enumerate(rest):
            section_html = _render_help_section(match, use_query_links=True)
            if idx < len(rest) - 1:
                section_html += '<hr class="help-sep">'
            sections.append(section_html)

        multi = f"<div class='help-multi'>{''.join(sections)}</div>"
        page = f"<h1>{title}</h1>{multi}"
        if "class='python'" in page:
            page += highlight_js
        return page

    # Not a multi-match result: just render normally
    body = _render_help_section(help_info, use_query_links=True)
    page = f"<h1>{title}</h1>{body}"
    if "class='python'" in page:
        page += highlight_js
    return page

def _render_help_section(info, use_query_links=False, highlight=False, *args, **kwargs):
    import html
    proj = info.get("Project")
    func = info.get("Function")
    header = ""
    if proj and func:
        if use_query_links:
            proj_link = f'<a href="?topic={proj}">{proj}</a>'
            func_link = f'<a href="?topic={proj}/{func}">{func}</a>'
        else:
            proj_link = html.escape(proj)
            func_link = html.escape(func)
        header = f"""
        <div class="projfunc-row">
            <span class="project">{proj_link}</span>
            <span class="dot">·</span>
            <span class="function">{func_link}</span>
        </div>
        """

    rows = []
    skip_keys = {"Project", "Function"}
    for key, value in info.items():
        if key in skip_keys:
            continue

        # 1. Only autolink References (and plain text fields).
        # 2. Don't autolink Sample CLI, Signature, Full Code, etc.

        if use_query_links and key == "References" and isinstance(value, (list, tuple)):
            refs = [
                f'<a href="?topic={ref}">{html.escape(str(ref))}</a>' for ref in value
            ]
            value = ', '.join(refs)
            value = f"<div class='refs'>{value}</div>"

        # Improvement 4: Copy to clipboard button for Full Code
        elif key == "Full Code":
            code_id = f"code_{abs(hash(value))}"
            value = (
                f"<div class='full-code-block'>"
                f"<button class='copy-btn' onclick=\"copyToClipboard('{code_id}')\">Copy to clipboard</button>"
                f"<pre><code id='{code_id}' class='python'>{html.escape(str(value))}</code></pre>"
                f"</div>"
                "<script>"
                "function copyToClipboard(codeId) {"
                "  var text = document.getElementById(codeId).innerText;"
                "  navigator.clipboard.writeText(text).then(()=>{"
                "    alert('Copied!');"
                "  });"
                "}"
                "</script>"
            )

        # Code fields: no autolinking, just escape & highlight
        elif key in ("Signature", "Example CLI", "Example Code", "Sample CLI"):
            value = f"<pre><code class='python'>{html.escape(str(value))}</code></pre>"

        elif key in ("Docstring", "TODOs"):
            value = f"<div class='doc'>{html.escape(str(value))}</div>"

        # Only for regular text fields, run _autolink_refs
        elif use_query_links and isinstance(value, str):
            value = _autolink_refs(value)
            value = f"<p>{value}</p>"

        else:
            value = f"<p>{html.escape(str(value))}</p>"

        rows.append(f"<section><h3>{key}</h3>{value}</section>")

    # Highlight exact matches with a CSS class
    article_class = 'help-entry'
    if highlight:
        article_class += ' help-entry-exact'
    return f"<article class='{article_class}'>{header}{''.join(rows)}</article>"

def _autolink_refs(text):
    import re
    return re.sub(r'\b([a-zA-Z0-9_]+)(?:\.([a-zA-Z0-9_]+))?\b', 
        lambda m: (
            f'<a href="?topic={m.group(1)}">{m.group(1)}</a>' if not m.group(2) 
            else f'<a href="?topic={m.group(1)}/{m.group(2)}">{m.group(1)}.{m.group(2)}</a>'
        ), text)

def view_qr_code(*args, value=None, **kwargs):
    """Generate a QR code for a given value and serve it from cache if available."""
    if not value:
        return '''
            <h1>QR Code Generator</h1>
            <form method="post">
                <input type="text" name="value" placeholder="Enter text or URL" required class="main" />
                <button type="submit" class="submit">Generate QR</button>
            </form>
        '''
    qr_url = gw.qr.generate_url(value)
    back_link = gw.web.app_url("qr-code")
    return f"""
        <h1>QR Code for:</h1>
        <h2><code>{value}</code></h2>
        <img src="{qr_url}" alt="QR Code" class="qr" />
        <p><a href="{back_link}">Generate another</a></p>
    """


def _create_github_issue(title: str, body: str) -> str:
    """Create an issue in the GWAY repository and return the issue URL."""
    return gw.hub.create_issue(title, body)


def view_feedback(*, name=None, email=None, topic=None, message=None, create_issue=None):
    """Display feedback form and submit feedback as a GitHub issue."""
    import html
    from bottle import request

    token = gw.hub.get_token()
    mail_configured = all(
        os.environ.get(k)
        for k in ["MAIL_SENDER", "MAIL_PASSWORD", "SMTP_SERVER", "SMTP_PORT"]
    )
    if request.method != "POST" and not token and not mail_configured:
        return (
            "<h1>Feedback unavailable</h1>"
            "<p>GitHub token and mail settings not configured.</p>"
        )

    if request.method == "POST":
        name = (name or "").strip()
        email = (email or "").strip()
        topic = (topic or "").strip()
        message = (message or "").strip()
        create_issue = bool(create_issue)

        missing = [field for field, val in [
            ("Name", name),
            ("Email", email),
            ("Topic", topic),
            ("Message", message),
        ] if not val]
        if missing:
            miss = ", ".join(missing)
            back = gw.web.app.build_url("feedback")
            return f"<h1>Missing required fields: {html.escape(miss)}</h1><p><a href='{back}'>Back</a></p>"

        issue_url = ""
        fallback_to_mail = False
        if create_issue:
            body = f"**From:** {name}\n\n{message}"
            try:
                issue_url = _create_github_issue(topic, body)
            except Exception:
                fallback_to_mail = True
                gw.mail.send(
                    f"Feedback: {topic}",
                    body=f"From: {name} <{email}>\n\n{message}",
                )
        else:
            gw.mail.send(
                f"Feedback: {topic}",
                body=f"From: {name} <{email}>\n\n{message}",
            )

        msg = "<h1>Thank you for your feedback!</h1>"
        if issue_url:
            msg += f"<p>It was recorded as <a href='{issue_url}'>GitHub issue</a>.</p>"
        elif fallback_to_mail:
            msg += "<p>GitHub issue creation failed; feedback sent via email.</p>"
        return msg

    issue_option = (
        "<label class=\"checkbox-right\">"
        "Optional: Create an Issue Report for GWAY or this website."
        "<input type=\"checkbox\" name=\"create_issue\" value=\"1\" />"
        "</label>"
    ) if token else "<p>(GitHub issue creation unavailable)</p>"
    return f"""
        <h1>Send Feedback</h1>
        <p>Your name and message may be publicly displayed and processed. Your email will be kept private.</p>
        <form method="post">
            <input type="text" name="name" placeholder="Your Name" required class="main" />
            <input type="email" name="email" placeholder="Email" required class="main" />
            <input type="text" name="topic" placeholder="Topic" required class="main" />
            <textarea name="message" placeholder="Message" required rows="6" class="main"></textarea>
            {issue_option}
            <button type="submit" class="submit btn-block">Submit</button>
        </form>
    """


def view_debug_info():
    """Return HTML with debug info about the current request and log tail."""
    from bottle import request
    import html as _html

    info = []
    info.append(f"<b>URL:</b> {_html.escape(request.url or '')}")
    info.append(f"<b>Method:</b> {_html.escape(request.method or '')}")
    info.append(f"<b>Version:</b> {_html.escape(gw.version())}")
    try:
        commit = gw.hub.commit()
        if commit:
            info.append(f"<b>Commit:</b> {_html.escape(commit)}")
    except Exception:
        pass

    log_tail = ""
    try:
        log_path = gw.resource("logs", "gway.log")
        with open(log_path, "r", encoding="utf-8") as lf:
            lines = lf.readlines()[-20:]
            log_tail = "".join(lines)
    except Exception:
        log_tail = "(log unavailable)"

    return (
        "<div class='debug-info'>" + "<br>".join(info) +
        f"<pre>{_html.escape(log_tail)}</pre></div>"
    )





import unittest
import subprocess
import time
import socket
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from gway import gw

class SiteHelpAutoTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.proc = subprocess.Popen(
            ["gway", "-r", "test/website"],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
        )
        cls._wait_for_port(18888, timeout=15)
        time.sleep(2)
        cls.base_url = "http://127.0.0.1:18888"

    @classmethod
    def tearDownClass(cls):
        if hasattr(cls, "proc") and cls.proc:
            cls.proc.terminate()
            try:
                cls.proc.wait(timeout=5)
            except subprocess.TimeoutExpired:
                cls.proc.kill()
        try:
            with gw.web.auto.browse(close=True):
                pass
        except Exception:
            pass

    @staticmethod
    def _wait_for_port(port, timeout=12):
        start = time.time()
        while time.time() - start < timeout:
            try:
                with socket.create_connection(("127.0.0.1", port), timeout=1):
                    return
            except OSError:
                time.sleep(0.2)
        raise TimeoutError(f"Port {port} not responding after {timeout} seconds")

    def test_help_search_finds_builtin(self):
        url = self.base_url + "/web/site/help"
        try:
            with gw.web.auto.browse(url=url) as drv:
                textarea = drv.find_element(By.ID, "help-search")
                textarea.clear()
                textarea.send_keys("hello-world")
                drv.find_element(By.CSS_SELECTOR, "form.nav").submit()
                WebDriverWait(drv, 10).until(
                    EC.text_to_be_present_in_element((By.TAG_NAME, "h1"), "hello_world")
                )
                h1_text = drv.find_element(By.TAG_NAME, "h1").text
                self.assertIn("hello_world", h1_text.lower())
        except Exception as e:
            self.skipTest(f"Webdriver unavailable: {e}")

    def test_search_box_autoexpands(self):
        """Search box should grow taller when text wraps to new lines."""
        url = self.base_url + "/web/site/help"
        long_text = "word " * 50
        try:
            with gw.web.auto.browse(url=url) as drv:
                textarea = drv.find_element(By.ID, "help-search")
                start_height = drv.execute_script(
                    "return arguments[0].clientHeight", textarea
                )
                textarea.clear()
                textarea.send_keys(long_text)
                # allow JS to process autoExpand
                WebDriverWait(drv, 5).until(
                    lambda d: d.execute_script(
                        "return arguments[0].clientHeight", textarea
                    )
                    > start_height
                )
                end_height = drv.execute_script(
                    "return arguments[0].clientHeight", textarea
                )
                self.assertGreater(
                    end_height,
                    start_height,
                    f"expected height to grow: {start_height} -> {end_height}",
                )
        except Exception as e:
            self.skipTest(f"Webdriver unavailable: {e}")

if __name__ == "__main__":
    unittest.main()




# projects/web/proxy.py

from fastapi import FastAPI
from gway import gw
import requests



def fallback_app(*,
        endpoint: str, app=None, websockets: bool = False, path: str = "/",
        mode: str = "extend", callback=None,
    ):
    """
    Create an HTTP (and optional WebSocket) fallback to the given endpoint.
    This assumes the remote endpoint replicates missing functionality or the
    entire service if it can't be provided locally.

    ``mode`` controls how the proxy is used:
    - ``replace``: forward all requests directly to the proxied endpoint.
    - ``extend``: redirect paths not configured locally to the proxy.
    - ``errors``: call the proxy only when the local handler raises an error or
      returns no data.
    - ``trigger``: call ``callback(request)``. Proxy if the callback returns
      ``True`` (supports async callbacks on FastAPI). ``callback`` must be
      provided for this mode.
    """
    # selectors for app types
    from bottle import Bottle



    # collect apps by type
    match app:
        case Bottle() as b:
            bottle_app, fastapi_app = b, None
        case FastAPI() as f:
            bottle_app, fastapi_app = None, f
        case list() | tuple() as seq:
            bottle_app = next((x for x in seq if isinstance(x, Bottle)), None)
            fastapi_app = next((x for x in seq if isinstance(x, FastAPI)), None)
        case None:
            bottle_app = fastapi_app = None
        case _ if isinstance(app, Bottle):
            bottle_app, fastapi_app = app, None
        case _ if isinstance(app, FastAPI):
            bottle_app, fastapi_app = None, app
        case _ if hasattr(app, "__iter__") and not isinstance(app, (str, bytes, bytearray)):
            bottle_app = next((x for x in app if isinstance(x, Bottle)), None)
            fastapi_app = next((x for x in app if isinstance(x, FastAPI)), None)
        case _:
            bottle_app = fastapi_app = None

    prepared = []
    
    # in replace mode, ignore existing routes and return proxy-only apps
    if mode == "replace":
        if bottle_app:
            new_bottle = Bottle()
            prepared.append(_wire_proxy(new_bottle, endpoint, websockets, path))
        if fastapi_app:
            new_fastapi = FastAPI()
            prepared.append(_wire_proxy(new_fastapi, endpoint, websockets, path))
        if not prepared:
            default = Bottle()
            prepared.append(_wire_proxy(default, endpoint, websockets, path))
    else:
        # if no matching apps, default to a new Bottle
        if not bottle_app and not fastapi_app:
            default = Bottle()
            prepared.append(_wire_proxy(default, endpoint, websockets, path))
            if mode == "errors":
                _wire_error_fallback(default, endpoint, path)
            elif mode == "trigger":
                _wire_trigger_fallback(default, endpoint, callback)
        elif bottle_app:
            prepared.append(_wire_proxy(bottle_app, endpoint, websockets, path))
            if mode == "errors":
                _wire_error_fallback(bottle_app, endpoint, path)
            elif mode == "trigger":
                _wire_trigger_fallback(bottle_app, endpoint, callback)
        elif fastapi_app:
            prepared.append(_wire_proxy(fastapi_app, endpoint, websockets, path))
            if mode == "errors":
                _wire_error_fallback(fastapi_app, endpoint, path)
            elif mode == "trigger":
                _wire_trigger_fallback(fastapi_app, endpoint, callback)

    return prepared[0] if len(prepared) == 1 else tuple(prepared)


def _wire_proxy(app, endpoint: str, websockets: bool, path: str):
    """
    Internal: attach HTTP and optional WS proxy routes
    to Bottle or FastAPI-compatible app. Both content and headers are proxied.
    """
    # detect FastAPI-like
    is_fastapi = hasattr(app, "websocket")

    # auto-enable websockets for FastAPI
    if is_fastapi and not websockets:
        websockets = True

    # FastAPI: new app if needed
    if app is None and websockets:
        from fastapi import FastAPI, WebSocket, WebSocketDisconnect, Request
        import httpx, websockets, asyncio

        app = FastAPI()
        base = path.rstrip("/") or "/"

        @app.api_route(f"{base}/{{full_path:path}}", methods=["GET","POST","PUT","PATCH","DELETE","OPTIONS","HEAD"])
        async def proxy_http(request: Request, full_path: str):
            url = endpoint.rstrip("/") + "/" + full_path
            client = httpx.AsyncClient()
            headers = dict(request.headers)
            body = await request.body()
            resp = await client.request(request.method, url, headers=headers, content=body)
            return resp.content, resp.status_code, resp.headers.items()

        @app.websocket(f"{base}/{{full_path:path}}")
        async def proxy_ws(ws: WebSocket, full_path: str):
            upstream = endpoint.rstrip("/") + "/" + full_path
            await ws.accept()
            try:
                async with websockets.connect(upstream) as up:
                    async def c2u():
                        while True:
                            m = await ws.receive_text()
                            await up.send(m)
                    async def u2c():
                        while True:
                            m = await up.recv()
                            await ws.send_text(m)
                    await asyncio.gather(c2u(), u2c())
            except WebSocketDisconnect:
                pass
            except Exception as e:
                gw.error(f"WebSocket proxy error: {e}")

        return app

    # Bottle-only HTTP proxy
    if hasattr(app, "route") and not is_fastapi:
        from bottle import request

        @app.route(f"{path}<path:path>", method=["GET","POST","PUT","DELETE","PATCH","OPTIONS","HEAD"])
        def _bottle_proxy(path):
            target = f"{endpoint.rstrip('/')}/{path}"
            headers = {k: v for k, v in request.headers.items()}
            try:
                resp = requests.request(request.method, target, headers=headers, data=request.body.read(), stream=True)
                return resp.content, resp.status_code, resp.headers.items()
            except Exception as e:
                gw.error("Proxy request failed: %s", e)
                return f"Proxy error: {e}", 502

        if websockets:
            gw.warning("WebSocket proxy requested but Bottle does not support WebSockets; ignoring websockets=True")

        return app

    # Existing FastAPI-like app augmentation
    if is_fastapi:
        from fastapi import WebSocket, WebSocketDisconnect, Request
        import httpx, websockets, asyncio

        base = path.rstrip("/") or "/"

        @app.api_route(f"{base}/{{full_path:path}}", methods=["GET","POST","PUT","PATCH","DELETE","OPTIONS","HEAD"])
        async def proxy_http(request: Request, full_path: str):
            url = endpoint.rstrip("/") + "/" + full_path
            client = httpx.AsyncClient()
            headers = dict(request.headers)
            body = await request.body()
            resp = await client.request(request.method, url, headers=headers, content=body)
            return resp.content, resp.status_code, resp.headers.items()

        if websockets:
            @app.websocket(f"{base}/{{full_path:path}}")
            async def proxy_ws(ws: WebSocket, full_path: str):
                upstream = endpoint.rstrip("/") + "/" + full_path
                await ws.accept()
                try:
                    async with websockets.connect(upstream) as up:
                        async def c2u():
                            while True:
                                m = await ws.receive_text()
                                await up.send(m)
                        async def u2c():
                            while True:
                                m = await up.recv()
                                await ws.send_text(m)
                        await asyncio.gather(c2u(), u2c())
                except WebSocketDisconnect:
                    pass
                except Exception as e:
                    gw.error(f"WebSocket proxy error: {e}")

        return app

    raise RuntimeError("Unsupported app type for fallback_app: must be Bottle or FastAPI-compatible")


def _wire_error_fallback(app, endpoint: str, path: str):
    """Attach error-handling hooks/middleware that proxies failed calls."""
    is_fastapi = hasattr(app, "websocket")

    if is_fastapi:
        from fastapi import Request, Response
        import httpx

        @app.middleware("http")
        async def _proxy_on_error(request: Request, call_next):
            body = await request.body()
            try:
                resp = await call_next(request)
                if resp.status_code >= 500 or resp.status_code == 404 or not getattr(resp, "body", b""):
                    raise Exception("fallback to proxy")
                return resp
            except Exception:
                url = endpoint.rstrip("/") + request.url.path
                if request.url.query:
                    url += "?" + request.url.query
                async with httpx.AsyncClient() as client:
                    proxied = await client.request(request.method, url, headers=dict(request.headers), content=body)
                return Response(content=proxied.content, status_code=proxied.status_code, headers=proxied.headers)
        return app

    if hasattr(app, "route"):
        from bottle import request, response

        def _do_proxy():
            target = endpoint.rstrip("/") + request.fullpath
            headers = {k: v for k, v in request.headers.items()}
            try:
                resp = requests.request(request.method, target, headers=headers, data=request.body.read(), stream=True)
                response.status = resp.status_code
                for k, v in resp.headers.items():
                    response.set_header(k, v)
                return resp.content
            except Exception as e:
                gw.error("Proxy request failed: %s", e)
                response.status = 502
                return f"Proxy error: {e}"

        @app.error(404)
        @app.error(500)
        def _handle_errors(err):
            return _do_proxy()

        @app.hook("after_request")
        def _after_req():
            if not response.body:
                response.body = _do_proxy()

        return app

    raise RuntimeError("Unsupported app type for fallback_app: must be Bottle or FastAPI-compatible")


def _wire_trigger_fallback(app, endpoint: str, callback):
    """Attach middleware/hooks to proxy when callback(request) returns True."""
    if not callable(callback):
        raise ValueError("callback must be callable for trigger mode")

    is_fastapi = hasattr(app, "websocket")

    if is_fastapi:
        from fastapi import Request, Response
        import httpx, inspect

        async def _do_proxy(request: Request) -> Response:
            url = endpoint.rstrip("/") + request.url.path
            if request.url.query:
                url += "?" + request.url.query
            async with httpx.AsyncClient() as client:
                proxied = await client.request(
                    request.method,
                    url,
                    headers=dict(request.headers),
                    content=await request.body(),
                )
            return Response(
                content=proxied.content,
                status_code=proxied.status_code,
                headers=proxied.headers,
            )

        @app.middleware("http")
        async def _proxy_on_trigger(request: Request, call_next):
            result = callback(request)
            if inspect.isawaitable(result):
                result = await result
            if result:
                return await _do_proxy(request)
            return await call_next(request)

        return app

    if hasattr(app, "route"):
        from bottle import request, HTTPResponse

        def _do_proxy():
            target = endpoint.rstrip("/") + request.fullpath
            headers = {k: v for k, v in request.headers.items()}
            try:
                resp = requests.request(
                    request.method,
                    target,
                    headers=headers,
                    data=request.body.read(),
                    stream=True,
                )
                return HTTPResponse(resp.content, status=resp.status_code, headers=resp.headers)
            except Exception as e:
                gw.error("Proxy request failed: %s", e)
                return HTTPResponse(f"Proxy error: {e}", status=502)

        @app.hook("before_request")
        def _before_req():
            if callback(request):
                raise _do_proxy()

        return app

    raise RuntimeError("Unsupported app type for fallback_app: must be Bottle or FastAPI-compatible")



import unittest
from pathlib import Path
from gway import gw


class FakeClient:
    def __init__(self, host):
        self.host = host


class FakeRequest:
    def __init__(self, addr):
        self.remote_addr = addr
        self.client = FakeClient(addr)


class IsLocalTests(unittest.TestCase):
    def test_local_address_returns_true(self):
        req = FakeRequest("127.0.0.1")
        self.assertTrue(gw.web.server.is_local(request=req, host="127.0.0.1"))

    def test_non_local_address_returns_false(self):
        req = FakeRequest("8.8.8.8")
        self.assertFalse(gw.web.server.is_local(request=req, host="127.0.0.1"))


if __name__ == "__main__":
    unittest.main()



# projects/gui.py

import sys
import platform
import subprocess
import re, glob, time, os
from datetime import datetime
import pygame
from PIL import Image, ImageGrab
from gway import gw


def notify(message, *, title="GWAY Notice", timeout: int=10):
    """
    Show a user interface notification with the specified title and message.
    Falls back to a visible console printout if GUI notification fails.
    """
    try:
        from plyer import notification
        notification.notify(
            title=title,
            message=message,
            app_name="gway",
            timeout=timeout
        )
        gw.info(f"Notification: {title} - {message}")
    except Exception as e:
        fallback = f"\n{'='*10} {title} {'='*10}\n{message}\n{'='*30}"
        print(fallback)
        gw.critical(f"Notification fallback: {fallback} (Error: {e})")


def instruct(steps, *, title="GWAY Next Steps", timeout: int=300, **kwargs):
    """Display a series of steps using a notify box."""
    split_steps = '\n'.join(steps).split('\n')
    num_steps = [f"{n+1}. {step}" for n, step in enumerate(steps) if steps.strip()]
    instructions = '\n'.join(num_steps)
    return notify(message=instructions, title=title, timeout=timeout, **kwargs)


def lookup_font(*prefix):
    """Look up fonts installed on a Windows system by partial name (prefix).
    >> gway font lookup Ari
    """
    import winreg
    font_prefix = " ".join(prefix)

    try:
        font_key_path = r"SOFTWARE\Microsoft\Windows NT\CurrentVersion\Fonts"
        with winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, font_key_path) as font_key:
            num_values = winreg.QueryInfoKey(font_key)[1]
            matching_fonts = []

            prefix_lower = font_prefix.lower()
            for i in range(num_values):
                value_name, value_data, _ = winreg.EnumValue(font_key, i)
                name_only = value_name.split(" (")[0].strip()

                if prefix_lower in name_only.lower() or prefix_lower in value_data.lower():
                    matching_fonts.append(f"{name_only} -> {value_data}")

            return matching_fonts if matching_fonts else [f"No match for prefix: {font_prefix}"]

    except Exception as e:
        return [f"Error: {str(e)}"]


def _get_active_window():
    system = platform.system()

    if system == "Windows":
        try:
            import ctypes
            user32 = ctypes.windll.user32
            hwnd = user32.GetForegroundWindow()
            length = user32.GetWindowTextLengthW(hwnd)
            buff = ctypes.create_unicode_buffer(length + 1)
            user32.GetWindowTextW(hwnd, buff, length + 1)
            return buff.value or "window"
        except Exception:
            return "window"

    elif system == "Darwin":
        try:
            script = 'tell application "System Events" to get name of first process whose frontmost is true'
            p = subprocess.Popen(
                ["osascript", "-e", script],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            out, _ = p.communicate(timeout=2)
            name = out.decode("utf-8").strip()
            return name if name else "window"
        except Exception:
            return "window"

    else:
        try:
            p = subprocess.Popen(
                ["xdotool", "getactivewindow", "getwindowname"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            out, _ = p.communicate(timeout=2)
            name = out.decode("utf-8").strip()
            return name if name else "window"
        except Exception:
            return "window"


def _sanitize_filename(name: str) -> str:
    cleaned = "".join(
        c if (c.isalnum() or c in (" ", "_", "-")) else "_" for c in name
    )
    return cleaned.replace(" ", "_").strip("_") or "window"



def shot(*, name: str = None, mode: str = "full") -> str:
    """
    Take a screenshot in the specified mode and save it under:
        gw.resource("work", "screenshots")

    The filename will be:
        <active_window_name>_<YYYYMMDD_HHMMSS>.png

    Returns:
        The full path to the saved screenshot file.

    Modes:
        - "full": entire screen
        - "active"/"window": active window only (Windows only; falls back to full)
    """

    screenshots_dir = gw.resource("work", "screenshots")
    os.makedirs(screenshots_dir, exist_ok=True)

    window_name = _get_active_window()
    window_name = _sanitize_filename(window_name)

    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    filename = f"{window_name}_{timestamp}.png"
    filepath = os.path.join(screenshots_dir, filename)

    if mode in ("active", "window"):
        try:
            import pygetwindow as gwnd
            win = gwnd.getActiveWindow()
            if win and win.left != -32000:  # Avoid minimized windows
                bbox = (win.left, win.top, win.right, win.bottom)
                img = ImageGrab.grab(bbox=bbox)
            else:
                img = ImageGrab.grab()
        except Exception:
            img = ImageGrab.grab()
    else:
        img = ImageGrab.grab()

    img.save(filepath)

    return filepath


def reminder(message, *, interval: float = 20.0, daemon=False, lines: int = 2):
    """
    Starts a thread that periodically takes screenshots.
    If the screen hasn't changed between intervals, overlays a reminder
    message and waits for user interaction before resuming.
    """
    import threading
    from PIL import ImageChops
    import pygame
    import ctypes

    def images_equal(img1, img2, threshold=5):
        diff = ImageChops.difference(img1.convert("L"), img2.convert("L"))
        bbox = diff.getbbox()
        if not bbox:
            return True
        stat = diff.crop(bbox).getextrema()
        return stat[1] < threshold

    def bring_to_front():
        if sys.platform == "win32":
            hwnd = pygame.display.get_wm_info()["window"]
            ctypes.windll.user32.SetWindowPos(hwnd, -1, 0, 0, 0, 0, 1 | 2)

    def get_screen_size():
        import tkinter as tk
        root = tk.Tk()
        root.withdraw()
        width = root.winfo_screenwidth()
        height = root.winfo_screenheight()
        root.destroy()
        return (width, height)

    def show_reminder(bg_img: Image.Image):
        import tkinter as tk
        from PIL import ImageTk

        screen_size = get_screen_size()
        reminder_img = render_text(message, size=screen_size, bg=bg_img, lines=lines)

        root = tk.Tk()
        root.title("Reminder")
        root.attributes("-fullscreen", True)
        root.configure(background='black')
        root.focus_force()

        img_tk = ImageTk.PhotoImage(reminder_img)
        label = tk.Label(root, image=img_tk)
        label.pack()

        root.bind("<Key>", lambda e: root.destroy())
        root.bind("<Button>", lambda e: root.destroy())
        root.bind("<Motion>", lambda e: root.destroy())

        root.mainloop()

    def loop():
        reminder_dir = gw.resource("work", "reminder")
        os.makedirs(reminder_dir, exist_ok=True)
        last_img = None

        while True:
            current = ImageGrab.grab()
            current.save(os.path.join(reminder_dir, "next.png"))

            if last_img and images_equal(current, last_img):
                current.save(os.path.join(reminder_dir, "original.png"))
                show_reminder(current)

            last_img = current
            time.sleep(float(interval))

    thread = threading.Thread(target=loop, daemon=daemon)
    thread.start()

    if not daemon:
        try:
            while thread.is_alive():
                thread.join(timeout=1)
        except (KeyboardInterrupt, EOFError):
            print("Exiting reminder.")
            return

    return thread


def render_text(text, *, size=(800, 600), bg=None, fg=(255, 255, 255), font_name=None, lines=2) -> Image.Image:
    import pygame
    import textwrap
    pygame.font.init()  # Only init what we need

    width, height = size
    MARGIN = 40

    # Natural wrapping
    wrapped = textwrap.wrap(text, width=30)[:int(lines)]

    img = pygame.Surface(size)
    if isinstance(bg, tuple):
        img.fill(bg)
    elif isinstance(bg, Image.Image):
        bg = bg.resize(size)
        bg_data = pygame.image.fromstring(bg.tobytes(), size, bg.mode)
        img.blit(bg_data, (0, 0))
    else:
        img.fill((0, 0, 0))

    font_size = 200
    while font_size > 10:
        font = pygame.font.SysFont(font_name, font_size)
        rendered = [font.render(line, True, fg) for line in wrapped]
        max_width = max((r.get_width() for r in rendered), default=0)
        total_height = sum(r.get_height() for r in rendered) + MARGIN * (len(rendered) - 1)

        if max_width <= width - 2 * MARGIN and total_height <= height - 2 * MARGIN:
            break
        font_size -= 2

    # Center vertically
    y = (height - total_height) // 2
    for r in rendered:
        x = (width - r.get_width()) // 2
        img.blit(r, (x, y))
        y += r.get_height() + MARGIN

    pygame.font.quit()  # Clean up

    return Image.frombytes("RGB", size, pygame.image.tostring(img, "RGB"))

...


def animate_gif(pattern, *, output_gif=None):
    resolved = gw.resource(pattern)
    if os.path.isdir(resolved):
        pngs = sorted(glob.glob(os.path.join(resolved, "*.png")))
        if not pngs:
            gw.abort(f"No .png files found in directory: {resolved}")
        sample_file = os.path.basename(pngs[0])
        base_dir = resolved
    else:
        sample_file = os.path.basename(resolved)
        base_dir = os.path.dirname(resolved) or "."

    # Detect [n] or “ - n” patterns (falls back to ctime sort)
    bracket = re.search(r'\[(\d+)\]', sample_file)
    dash    = re.search(r'^(.*?)([ \-_]+)(\d+)(\.png)$', sample_file)
    if bracket:
        pfx = sample_file.split(bracket.group(0))[0]
        sfx = sample_file.split(bracket.group(0))[1]
        rx  = re.compile(r'^' + re.escape(pfx) + r'\[(\d+)\]' + re.escape(sfx) + r'$')
        pat = f"{pfx}*{sfx}"
    elif dash:
        pfx = dash.group(1) + dash.group(2)
        sfx = dash.group(4)
        rx  = re.compile(r'^' + re.escape(pfx) + r'(\d+)' + re.escape(sfx) + r'$')
        pat = f"{pfx}*{sfx}"
    else:
        # No numbering → creation order
        fns = sorted(glob.glob(os.path.join(base_dir, "*.png")), key=os.path.getctime)
        if not fns:
            gw.abort(f"No .png files in {base_dir}")
        images = [Image.open(f).convert("RGBA") for f in fns]
        return _display_and_save(images, fns, _make_outpath(pattern, output_gif, base_dir))

    # Gather & sort numbered frames
    items = []
    for fn in glob.glob(os.path.join(base_dir, pat)):
        nm = os.path.basename(fn)
        m  = rx.match(nm)
        if m: items.append((int(m.group(1)), fn))
    if not items:
        gw.abort(f"No files matching pattern {pat!r}")
    items.sort(key=lambda x: x[0])
    fns = [fn for _, fn in items]
    images = [Image.open(fn).convert("RGBA") for fn in fns]

    return _display_and_save(images, fns, _make_outpath(pattern, output_gif, base_dir))


def _make_outpath(pattern, output_gif, base_dir):
    if output_gif:
        return output_gif
    base = os.path.basename(pattern.rstrip("/\\")) or "output"
    return os.path.join(base_dir, base + ".gif")


def _display_and_save(pil_images, frame_files, output_gif):
    # Helper: flatten transparent frames onto black
    def flatten_rgba(img):
        if img.mode != "RGBA":
            return img.convert("RGB")
        bg = Image.new("RGB", img.size, (0, 0, 0))
        bg.paste(img, mask=img.split()[3])
        return bg

    # 1) Show & time
    pygame.init()
    W, H = zip(*(im.size for im in pil_images))
    screen = pygame.display.set_mode((max(W), max(H)))
    pygame.display.set_caption("SPACE to advance")

    durations, last = [], None
    font = pygame.font.SysFont(None, 36)
    for i, img in enumerate(pil_images):
        surf = pygame.image.fromstring(img.tobytes(), img.size, img.mode)
        screen.fill((0,0,0)); screen.blit(surf,(0,0))
        screen.blit(font.render(f"Frame {i}",True,(255,255,255)),(10,10))
        pygame.display.flip()
        waiting = True
        while waiting:
            for e in pygame.event.get():
                if e.type==pygame.KEYDOWN and e.key==pygame.K_SPACE:
                    now = time.time()
                    if last is not None: durations.append(now-last)
                    last = now
                    waiting = False
                elif e.type==pygame.QUIT:
                    pygame.quit(); gw.abort("User closed window")
    pygame.quit()
    if len(durations)==len(pil_images)-1: durations.append(durations[-1])
    durations_ms = [int(d*1000) for d in durations]

    # 2) Flatten + global quantize
    flat = [flatten_rgba(im) for im in pil_images]
    master = flat[0].convert("P", palette=Image.ADAPTIVE, colors=256)
    pal   = [im.quantize(palette=master) for im in flat]

    # 3) Save full-frame, browser-compatible GIF
    pal[0].save(
        output_gif,
        save_all=True,
        append_images=pal[1:],
        duration=durations_ms,
        loop=0,
        disposal=1,
        optimize=True
    )
    print(f"Saved GIF → {output_gif}")




# file: projects/ocpp/data.py
"""OCPP data helpers using gw.sql for storage."""

import time
import json
from datetime import datetime
from typing import Iterable, Optional, Sequence
from gway import gw

_db_conn = None

def open_db():
    """Return connection to the OCPP database, initializing tables."""
    global _db_conn
    if _db_conn is None:
        _db_conn = gw.sql.open_connection("work/ocpp.sqlite")
        _init_db(_db_conn)
    return _db_conn

def _init_db(conn):
    gw.sql.execute(
        """
        CREATE TABLE IF NOT EXISTS transactions(
            charger_id TEXT,
            transaction_id INTEGER,
            start_time INTEGER,
            stop_time INTEGER,
            id_tag TEXT,
            meter_start REAL,
            meter_stop REAL,
            reason TEXT,
            charger_start_ts INTEGER,
            charger_stop_ts INTEGER
        )
        """,
        connection=conn,
    )
    gw.sql.execute(
        """
        CREATE TABLE IF NOT EXISTS meter_values(
            charger_id TEXT,
            transaction_id INTEGER,
            timestamp INTEGER,
            measurand TEXT,
            value REAL,
            unit TEXT,
            context TEXT
        )
        """,
        connection=conn,
    )
    gw.sql.execute(
        """
        CREATE TABLE IF NOT EXISTS errors(
            charger_id TEXT,
            status TEXT,
            error_code TEXT,
            info TEXT,
            timestamp INTEGER
        )
        """,
        connection=conn,
    )

def record_transaction_start(
    charger_id: str,
    transaction_id: int,
    start_time: int,
    *,
    id_tag: Optional[str] = None,
    meter_start: Optional[float] = None,
    charger_timestamp: Optional[int] = None,
):
    conn = open_db()
    gw.sql.execute(
        "INSERT INTO transactions(charger_id, transaction_id, start_time, id_tag, meter_start, charger_start_ts) VALUES (?,?,?,?,?,?)",
        connection=conn,
        args=(
            charger_id,
            transaction_id,
            int(start_time),
            id_tag,
            meter_start,
            charger_timestamp,
        ),
    )

def record_transaction_stop(
    charger_id: str,
    transaction_id: int,
    stop_time: int,
    *,
    meter_stop: Optional[float] = None,
    reason: Optional[str] = None,
    charger_timestamp: Optional[int] = None,
):
    conn = open_db()
    gw.sql.execute(
        "UPDATE transactions SET stop_time=?, meter_stop=?, reason=?, charger_stop_ts=? WHERE charger_id=? AND transaction_id=?",
        connection=conn,
        args=(
            int(stop_time),
            meter_stop,
            reason,
            charger_timestamp,
            charger_id,
            transaction_id,
        ),
    )

def record_meter_value(charger_id: str, transaction_id: int, timestamp: int, measurand: str, value: float, unit: str = "", context: str = ""):
    conn = open_db()
    gw.sql.execute(
        "INSERT INTO meter_values(charger_id, transaction_id, timestamp, measurand, value, unit, context) VALUES (?,?,?,?,?,?,?)",
        connection=conn,
        args=(charger_id, transaction_id, int(timestamp), measurand, float(value), unit, context),
    )

def record_error(charger_id: str, status: str, error_code: str = "", info: str = ""):
    conn = open_db()
    ts = int(time.time())
    gw.sql.execute(
        "INSERT INTO errors(charger_id, status, error_code, info, timestamp) VALUES (?,?,?,?,?)",
        connection=conn,
        args=(charger_id, status, error_code, info, ts),
    )

def get_summary():
    """Return summary rows per charger."""
    conn = open_db()
    rows = gw.sql.execute(
        """
        SELECT t.charger_id AS cid,
               COUNT(t.transaction_id) AS sessions,
               SUM(COALESCE(t.meter_stop,0) - COALESCE(t.meter_start,0)) AS energy,
               MAX(t.stop_time) AS last_stop,
               (
                 SELECT e.error_code FROM errors e
                 WHERE e.charger_id=t.charger_id
                 ORDER BY e.timestamp DESC LIMIT 1
               ) AS last_error
        FROM transactions t
        GROUP BY t.charger_id
        """,
        connection=conn,
    )
    summary = []
    for cid, sessions, energy, last_stop, last_error in rows:
        summary.append({
            "charger_id": cid,
            "sessions": sessions,
            "energy": round((energy or 0.0) / 1000.0, 3),
            "last_stop": last_stop,
            "last_error": last_error,
        })
    return summary

def _fmt_time(ts: Optional[int]) -> str:
    if not ts:
        return "-"
    try:
        return datetime.utcfromtimestamp(int(ts)).isoformat() + "Z"
    except Exception:
        return str(ts)

def _parse_date(date_str: str) -> Optional[int]:
    """Parse YYYY-MM-DD or ISO datetime to epoch seconds."""
    if not date_str:
        return None
    try:
        if len(date_str) == 10:
            dt = datetime.fromisoformat(date_str)
        else:
            dt = datetime.fromisoformat(date_str.replace("Z", ""))
        return int(dt.timestamp())
    except Exception:
        return None

def iter_transactions(
    charger_id: str = None,
    *,
    start: int = None,
    end: int = None,
    sort: str = "start_time",
    order: str = "desc",
    limit: int = 50,
    offset: int = 0,
) -> Iterable[tuple]:
    """Iterate transaction rows with optional filtering and sorting."""
    conn = open_db()
    sql = (
        "SELECT charger_id, transaction_id, start_time, stop_time, meter_start, meter_stop, reason, id_tag "
        "FROM transactions WHERE 1=1"
    )
    args: list = []
    if charger_id:
        sql += " AND charger_id=?"
        args.append(charger_id)
    if start:
        sql += " AND start_time>=?"
        args.append(int(start))
    if end:
        sql += " AND start_time<=?"
        args.append(int(end))
    valid = {"start_time", "stop_time", "meter_start", "meter_stop"}
    if sort not in valid:
        sort = "start_time"
    order_sql = "DESC" if str(order).lower() != "asc" else "ASC"
    sql += f" ORDER BY {sort} {order_sql}"
    sql += " LIMIT ? OFFSET ?"
    args.extend([int(limit), int(offset)])
    return gw.sql.execute(sql, connection=conn, args=tuple(args))

def get_meter_series(chargers: Sequence[str], *, start: int = None, end: int = None):
    """Return dict of charger_id -> list of (timestamp, kWh)."""
    conn = open_db()
    data = {}
    for cid in chargers:
        sql = (
            "SELECT timestamp, value FROM meter_values "
            "WHERE charger_id=? AND measurand='Energy.Active.Import.Register'"
        )
        args = [cid]
        if start:
            sql += " AND timestamp>=?"
            args.append(int(start))
        if end:
            sql += " AND timestamp<=?"
            args.append(int(end))
        sql += " ORDER BY timestamp"
        rows = gw.sql.execute(sql, connection=conn, args=tuple(args))
        data[cid] = [(int(ts), float(val)) for ts, val in rows]
    return data

def list_chargers() -> list[str]:
    """Return list of distinct charger_ids."""
    conn = open_db()
    rows = gw.sql.execute(
        "SELECT DISTINCT charger_id FROM transactions ORDER BY charger_id",
        connection=conn,
    )
    return [r[0] for r in rows]

def view_charger_summary(**_):
    """Simple HTML summary of charger data."""
    rows = get_summary()
    html = [
        '<link rel="stylesheet" href="/static/ocpp/csms/charger_status.css">',
        "<h1>OCPP Charger Summary</h1>",
    ]
    if not rows:
        html.append("<p>No data.</p>")
        return "\n".join(html)
    html.append("<table class='ocpp-summary'>")
    html.append(
        "<tr><th>Charger</th><th>Sessions</th><th>Energy(kWh)</th><th>Last Stop</th><th>Last Error</th></tr>"
    )
    for r in rows:
        html.append(
            f"<tr><td>{r['charger_id']}</td><td>{r['sessions']}</td><td>{r['energy']}</td>"
            f"<td>{_fmt_time(r['last_stop'])}</td><td>{r['last_error'] or '-'}" + "</td></tr>"
        )
    html.append("</table>")
    return "\n".join(html)

def view_charger_details(
    *,
    charger_id: str = None,
    page: int = 1,
    sort: str = "start_time",
    order: str = "desc",
    since: str = None,
    until: str = None,
    **_,
):
    """Paginated table of transactions for a specific charger."""
    page = int(page or 1)
    offset = (page - 1) * 50
    start_ts = _parse_date(since)
    end_ts = _parse_date(until)
    chargers = list_chargers()
    if not charger_id and chargers:
        charger_id = chargers[0]
    rows = list(
        iter_transactions(
            charger_id,
            start=start_ts,
            end=end_ts,
            sort=sort,
            order=order,
            limit=50,
            offset=offset,
        )
    )
    html = [f"<h1>Transactions for {charger_id or 'All'}</h1>"]
    html.append("<form method='get' style='margin-bottom:1em;'>")
    html.append("<label>Charger: <select name='charger_id'>")
    for cid in chargers:
        sel = " selected" if cid == charger_id else ""
        html.append(f"<option value='{cid}'{sel}>{cid}</option>")
    html.append("</select></label> ")
    html.append(
        f"<label>Since: <input type='date' name='since' value='{since or ''}'></label> "
    )
    html.append(
        f"<label>Until: <input type='date' name='until' value='{until or ''}'></label> "
    )
    html.append("<label>Sort: <select name='sort'>")
    for f in ["start_time", "stop_time", "meter_start", "meter_stop"]:
        sel = " selected" if f == sort else ""
        html.append(f"<option value='{f}'{sel}>{f}</option>")
    html.append("</select></label> ")
    html.append("<label>Order: <select name='order'>")
    for o in ["asc", "desc"]:
        sel = " selected" if o == order else ""
        html.append(f"<option value='{o}'{sel}>{o}</option>")
    html.append("</select></label> ")
    html.append("<button type='submit'>Apply</button></form>")
    html.append("<table class='ocpp-details'>")
    html.append(
        "<tr><th>ID</th><th>Start</th><th>Stop</th><th>Meter Δ(kWh)</th><th>Reason</th></tr>"
    )
    for r in rows:
        delta = (r[5] or 0) - (r[4] or 0)
        html.append(
            f"<tr><td>{r[1]}</td><td>{_fmt_time(r[2])}</td><td>{_fmt_time(r[3])}</td>"
            f"<td>{round(delta/1000.0,3)}</td><td>{r[6] or ''}</td></tr>"
        )
    html.append("</table>")
    next_page = page + 1 if len(rows) >= 50 else None
    if page > 1 or next_page:
        html.append("<div class='pager'>")
        if page > 1:
            html.append(
                f"<a href='?charger_id={charger_id}&page={page-1}&sort={sort}&order={order}'>Prev</a>"
            )
        if next_page:
            html.append(
                f" <a href='?charger_id={charger_id}&page={next_page}&sort={sort}&order={order}'>Next</a>"
            )
        html.append("</div>")
    return "\n".join(html)

def view_time_series(*, chargers: list = None, start: str = None, end: str = None, **_):
    """Graph of energy usage over time for selected chargers."""
    chargers_all = list_chargers()
    chargers = gw.cast.to_list(chargers) if chargers else chargers_all
    start_ts = _parse_date(start)
    end_ts = _parse_date(end)
    series = get_meter_series(chargers, start=start_ts, end=end_ts)
    html = ["<h1>Energy Time Series</h1>"]
    html.append("<form method='get' style='margin-bottom:1em;'>")
    for cid in chargers_all:
        checked = "checked" if cid in chargers else ""
        html.append(
            f"<label style='margin-right:.6em;'><input type='checkbox' name='chargers' value='{cid}' {checked}> {cid}</label>"
        )
    html.append(
        f"<label>Start: <input type='date' name='start' value='{start or ''}'></label> "
    )
    html.append(
        f"<label>End: <input type='date' name='end' value='{end or ''}'></label> "
    )
    html.append("<button type='submit'>Show</button></form>")
    html.append('<script src="https://cdn.jsdelivr.net/npm/chart.js"></script>')
    html.append('<canvas id="tschart" height="320"></canvas>')
    html.append('<script>')
    html.append(f"const series = {json.dumps(series)};")
    html.append(
        "const datasets = Object.entries(series).map(([cid, vals]) => ({label: cid, data: vals.map(v => ({x:v[0]*1000,y:v[1]}))}));"
    )
    html.append(
        "new Chart(document.getElementById('tschart'), {type:'line', data:{datasets}, options:{parsing:false, scales:{x:{type:'time',time:{unit:'day'}}, y:{title:{display:true,text:'kWh'}}}}});"
    )
    html.append('</script>')
    return "\n".join(html)




import unittest
import tempfile
import os
from pathlib import Path
from gway import gw

class ViewChangelogTests(unittest.TestCase):
    def setUp(self):
        self.tmp = tempfile.TemporaryDirectory()
        self.base = Path(self.tmp.name)
        self.old_cwd = Path.cwd()
        os.chdir(self.base)

    def tearDown(self):
        os.chdir(self.old_cwd)
        self.tmp.cleanup()

    def _write_changelog(self, text: str):
        Path("CHANGELOG.rst").write_text(text)

    def test_header_hidden_when_empty(self):
        content = """Changelog
=========

Unreleased
----------

0.1.0 [build 123]
-----------------

- first change
"""
        self._write_changelog(content)
        html = gw.release.view_changelog()
        self.assertNotIn("Unreleased", html)

    def test_header_shown_when_has_entries(self):
        content = """Changelog
=========

Unreleased
----------
- new feature

0.1.0 [build 123]
-----------------

- first change
"""
        self._write_changelog(content)
        html = gw.release.view_changelog()
        self.assertIn("Unreleased", html)


if __name__ == "__main__":
    unittest.main()



import unittest
import importlib.util
from pathlib import Path
from unittest.mock import patch, MagicMock

class RpiCloneTests(unittest.TestCase):
    @staticmethod
    def _load_rpi():
        rpi_path = Path(__file__).resolve().parents[1] / 'projects' / 'rpi.py'
        spec = importlib.util.spec_from_file_location('rpi', rpi_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    @classmethod
    def setUpClass(cls):
        cls.rpi = cls._load_rpi()

    def test_ru_invokes_dd(self):
        with patch.object(self.rpi.subprocess, 'run') as run_mock:
            run_mock.return_value = MagicMock(returncode=0, stdout='ok', stderr='')
            result = self.rpi.ru('/dev/test')
            run_mock.assert_called_with([
                'sudo', 'dd', 'if=/dev/mmcblk0', 'of=/dev/test', 'bs=4M',
                'status=progress', 'conv=fsync'
            ], capture_output=True, text=True)
            self.assertEqual(result, 'ok')

    def test_view_pi_remote_lists_devices(self):
        with patch.object(self.rpi, '_list_devices', return_value=['/dev/sda']), \
             patch.object(self.rpi.gw.web.app, 'render_template', lambda **kw: kw['content']):
            html = self.rpi.view_pi_remote()
        self.assertIn('/dev/sda', html)
        self.assertIn('clone-progress', html)

    def test_view_pi_remote_starts_clone(self):
        def fake_thread(target, args=(), daemon=None):
            class _T:
                def start(self):
                    target(*args)
            return _T()

        with patch.object(self.rpi, '_list_devices', return_value=['/dev/sda']), \
             patch.object(self.rpi.gw.web.app, 'render_template', lambda **kw: kw['content']), \
             patch.object(self.rpi, 'ru') as ru_mock, \
             patch.object(self.rpi.threading, 'Thread', side_effect=fake_thread):
            html = self.rpi.view_pi_remote(target='/dev/sda')
        ru_mock.assert_called_with('/dev/sda')
        self.assertIn('clone-progress', html)
        self.assertEqual(self.rpi._CLONE_STATE['progress'], 100.0)

if __name__ == '__main__':
    unittest.main()
